"""
Plugin toclusters.py:
    Implements inter-pump routing by filtering destinations.
    This is placed on a sarra process copying data between pumps.
    Whenever it gets a message, it looks at the destination and processing
    only continues if it is beleived that that message is a valid destination for the local pump.

Options:
    The local pump will select messages destined for the DD or DDI clusters, and reject those for DDSR, which isn't in the list.
        - msgToClusters DDI
        - msgToClusters DD
Usage:
    flowcb sarracenia.flowcb.accept.toclusters.ToClusters
    msgToClusters x
    msgToClusters y
    ...
"""

import os, stat, time
import logging
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class ToClusters(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

        self.o.add_option('msgToClusters', 'list')

        if self.o.msgToClusters == None:
            logger.info("msgToClusters setting mandatory")
            return

        logger.info("msgToClusters valid destinations: %s " % self.o.msgToClusters)

    def after_accept(self, worklist):
        new_incoming = []
        for message in worklist.incoming:
            if message['to_clusters'] in self.o.msgToClusters:
                new_incoming.append(message)
            else:
                worklist.rejected.append(message)
        worklist.incoming = new_incoming



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2020
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
import copy
from base64 import b64decode, b64encode
from collections import *
from hashlib import sha512
import json
import logging
from mimetypes import guess_type
import os
import os.path
import random
from random import choice

import sarracenia
from sarracenia import *
from sarracenia.featuredetection import features
from sarracenia.flowcb import FlowCB
import sarracenia.identity


import stat
from sys import platform as _platform
import sys
import time

if features['watch']['present']:
    from watchdog.observers import Observer
    from watchdog.observers.polling import PollingObserver
    from watchdog.events import PatternMatchingEventHandler

logger = logging.getLogger(__name__)


if features['watch']['present']:
    class SimpleEventHandler(PatternMatchingEventHandler):
        def __init__(self, parent):
            self.on_created = parent.on_created
            self.on_deleted = parent.on_deleted
            self.on_modified = parent.on_modified
            self.on_moved = parent.on_moved
            super().__init__()

def path_inflight_tooNew(inflight, lstat):
    """
      check the inflight, compare fail age against it.
      return True if the file is too new to be posted.
    """

    if not isinstance(inflight, int):
        #logger.debug("ok inflight unused")
        return False

    if lstat == None or not hasattr(lstat,'st_mtime'):
        #logger.debug("ok lstat None")
        return False

    age = nowflt() - lstat.st_mtime
    if age < inflight:
        logger.debug("%d vs (inflight setting) %d seconds. Too New!" % \
            (age,inflight) )
        return True

    return False




class File(FlowCB):
    """
    read the file system, create messages for the files you find.

    this is taken from v2's sr_post.py

    FIXME FIXME FIXME
    FIXME: the sr_post version would post files on the fly as it was traversing trees.
    so it was incremental and had little overhead. This one is does the whole recursion
    in one gather.

    It will fail horribly for large trees. Need to re-formulate to replace recursion with interation.
    perhaps a good time to use python iterators.
  
    also should likely switch from listdir to scandir
    """
    def on_add(self, event, src, dst):
        logger.debug("on_add %s %s %s" % ( event, src, dst ) )
        self.new_events['%s %s' % (src, dst)] = (event, src, dst)

    def on_created(self, event):
        # on_created (for SimpleEventHandler)
        if event.is_directory:
            self.on_add('mkdir', event.src_path, None)
        else:
            self.on_add('create', event.src_path, None)

    def on_deleted(self, event):
        # on_deleted (for SimpleEventHandler)
        if event.is_directory:
            self.on_add('rmdir', event.src_path, None)
        else:
            self.on_add('delete', event.src_path, None)

    def on_modified(self, event):
        # on_modified (for SimpleEventHandler)
        if not event.is_directory:
            self.on_add('modify', event.src_path, None)

    def on_moved(self, event):
        # on_moved (for SimpleEventHandler)
        self.on_add('move', event.src_path, event.dest_path)

    def __init__(self, options):
        """ 
        """

        super().__init__(options,logger)

        if not features['watch']['present']:
            logger.critical("watchdog module must be installed to watch directories")
            
        logger.debug("%s used to be overwrite_defaults" % self.o.component)

        self.obs_watched = []
        self.watch_handler = None
        self.post_topicPrefix = ["v03"]

        self.inl = OrderedDict()
        self.new_events = OrderedDict()
        self.left_events = OrderedDict()

        self.o.blocksize = 200 * 1024 * 1024
        self.o.create_modify = ('create' in self.o.fileEvents) or (
            'modify' in self.o.fileEvents)


    def post_delete(self, path, key=None, value=None,is_directory=False):
        #logger.debug("post_delete %s (%s,%s)" % (path, key, value))

        msg = sarracenia.Message.fromFileInfo(path, self.o, None)

        msg['fileOp'] = { 'remove':'' }

        if is_directory: 
            msg['fileOp']['directory'] = ''

        # partstr
        partstr = None

        # used when moving a file
        if key != None:
            msg[key] = value
            if key == 'newname' and self.o.post_baseDir:
                msg['new_dir'] = os.path.dirname(value)
                msg['new_file'] = os.path.basename(value)
                msg[key] = value.replace(self.o.post_baseDir, '')

        return [msg]

    def post_file(self, path, lstat, key=None, value=None):
        #logger.debug("post_file %s" % path)

        # check if it is a part file
        if path.endswith('.' + self.o.part_ext):
            return self.post_file_part(path, lstat)

        # This variable means that part_file_assemble plugin is loaded and will handle posting the original file (being assembled)

        elif hasattr(self, 'suppress_posting_partial_assembled_file'):
            return []

        # check the value of blocksize

        fsiz = lstat.st_size
        blksz = self.set_blocksize(self.o.blocksize, fsiz)

        # if we should send the file in parts

        #if blksz > 0 and blksz < fsiz:
        #    return self.post_file_in_parts(path, lstat)

        msg = sarracenia.Message.fromFileData(path, self.o, lstat)

        # used when moving a file
        if key != None:
            if not 'fileOp' in msg:
                msg['fileOp'] = { key : value }
            else:
                msg['fileOp'][key] = value

        if os_stat.S_ISDIR(lstat.st_mode):
            return [msg]

        # complete message
        if (self.o.post_topicPrefix[0] == 'v03') and self.o.inline:
            if fsiz < self.o.inlineByteMax:

                if self.o.inlineEncoding == 'guess':
                    e = guess_type(path)[0]
                    binary = not e or not ('text' in e)
                else:
                    binary = (self.o.inlineEncoding == 'text')

                f = open(path, 'rb')
                d = f.read()
                f.close()

                if binary:
                    msg["content"] = {
                        "encoding": "base64",
                        "value": b64encode(d).decode('utf-8')
                    }
                else:
                    try:
                        msg["content"] = {
                            "encoding": "utf-8",
                            "value": d.decode('utf-8')
                        }
                    except:
                        msg["content"] = {
                            "encoding": "base64",
                            "value": b64encode(d).decode('utf-8')
                        }
            else:
                if self.o.inlineOnly:
                    logger.error('skipping file %s too large (%d bytes > %d bytes max)) for inlining' % \
                       ( path, fsiz, self.o.inlineByteMax )  )
                    return []

        return [msg]

    def post_file_in_parts(self, path, lstat):
        #logger.debug("post_file_in_parts %s" % path )

        msg = sarracenia.Message.fromFileInfo(path, self.o, lstat)

        # check the value of blocksize

        fsiz = lstat.st_size
        chunksize = self.set_blocksize(self.o.blocksize, fsiz)

        # count blocks and remainder

        block_count = int(fsiz / chunksize)
        remainder = fsiz % chunksize
        if remainder > 0: block_count = block_count + 1

        # default sumstr

        sumstr = self.o.sumflg

        # loop on chunks

        blocks = list(range(0, block_count))
        if self.randomize:
            random.shuffle(blocks)
            #blocks = [8, 3, 1, 2, 9, 6, 0, 7, 4, 5] # Testing
            logger.info('Sending partitions in the following order: ' +
                        str(blocks))

        messages = []
        for i in blocks:

            # setting sumalgo for that part

            sumflg = self.o.sumflg

            if sumflg[:2] == 'z,' and len(sumflg) > 2:
                sumstr = sumflg

            else:
                sumflg = self.o.sumflg
                sumalgo = sarracenia.identity.Identity.factory(sumflg)
                sumalgo.set_path(path)

            # compute block stuff

            current_block = i

            offset = current_block * chunksize
            length = chunksize

            last = current_block == block_count - 1
            if last and remainder > 0:
                length = remainder

            # set partstr

            partstr = {
                'method': 'inplace',
                'size': chunksize,
                'count': block_count,
                'remainder': remainder,
                'number': current_block
            }
            # compute checksum if needed

            if not self.sumflg in ['random', 'cod']:
                bufsize = self.o.bufsize
                if length < bufsize: bufsize = length

                fp = open(path, 'rb')
                if offset != 0: fp.seek(offset, 0)
                t = 0
                while t < length:
                    buf = fp.read(bufsize)
                    if not buf: break
                    sumalgo.update(buf)
                    t += len(buf)
                fp.close()

                checksum = sumalgo.value
                sumstr = {'method': sumflg, 'value': checksum}

            # complete  message

            msg['identity'] = sumstr
            messages.extend(copy.deepcopy(msg))

        return messages

    def post_file_part(self, path, lstat):

        msg = sarracenia.Message.fromFileInfo(path, self.o, lstat)

        # verify suffix

        ok, log_msg, suffix, partstr, sumstr = self.msg.verify_part_suffix(
            path)

        # something went wrong

        if not ok:
            logger.debug("file part extension but %s for file %s" %
                         (log_msg, path))
            return False

        # check rename see if it has the right part suffix (if present)
        if 'rename' in self.msg.headers and not suffix in self.msg.headers[
                'rename']:
            msg['rename'] += suffix

        # complete  message

        msg['parts'] = partstr
        msg['identity'] = sumstr

        return [msg]

    def post_link(self, path, key='link', value=None):
        #logger.debug("post_link %s" % path )

        msg = sarracenia.Message.fromFileInfo(path, self.o, None)

        # resolve link

        if key == 'link':
            value = os.readlink(path)

        # used when moving a file
        if not 'fileOp' in msg:
           msg['fileOp'] = { key: value }
        else:
           msg['fileOp'][key] = value

        return [msg]

    def post_move(self, src, dst):
        #logger.debug("post_move %s %s" % (src,dst) )

        # watchdog funny ./ added at end of directory path ... removed

        messages = []
        src = src.replace('/./', '/')
        dst = dst.replace('/./', '/')

        if os.path.islink(dst) and self.o.realpathPost:
            dst = os.path.realpath(dst)
            if sys.platform == 'win32':
                dst = dst.replace('\\', '/')

        # file

        if os.path.isfile(dst):
            if hasattr(self.o,'v2compatRenameDoublePost') and self.o.v2compatRenameDoublePost:
                messages.extend(self.post_delete(src, 'newname', dst))
            messages.extend(self.post_file(dst, sarracenia.stat(dst), 'rename', src))
            return messages

        # link

        if os.path.islink(dst):
            if hasattr(self.o,'v2compatRenameDoublePost') and self.o.v2compatRenameDoublePost:
                messages.extend(self.post_delete(src, 'newname', dst))
            messages.extend(self.post_link(dst, 'rename', src))
            return messages

        # directory
        if os.path.isdir(dst):
            for x in os.listdir(dst):

                dst_x = dst + '/' + x
                src_x = src + '/' + x

                messages = self.post_move(src_x, dst_x)

            # directory list to delete at end
            self.move_dir_lst.append((src, dst))

        return messages

    def post1file(self, path, lstat, is_directory=False) -> list:
        """
          create the notification message for a single file, based on the lstat metadata.

          when lstat is present it is used to decide whether the file is an ordinary file, a link
          or a directory, and the appropriate message is built and returned.

          if the lstat metadata is None, then that signifies a "remove" message to be created.
          In the remove case, without the lstat, one needs the is_directory flag to decide whether
          it is an ordinary file remove, or a directory remove.   is_directory is not used other
          than for the remove case.

          The return value is a list that usually contains a single message.  It is a list to allow
          for if options are combined such that a symbolic link and the realpath it posts to may
          involve multiple messages for a single file. Similarly in the multi-block transfer case.

        """

        messages = []

        # watchdog funny ./ added at end of directory path ... removed
        path = path.replace('/./', '/')

        # always use / as separator for paths being posted.
        if os.sep != '/':  # windows
            path = path.replace(os.sep, '/')

        # path is a link

        if os.path.islink(path):
            messages.extend(self.post_link(path))

            if self.o.follow_symlinks:
                link = os.readlink(path)
                try:
                    rpath = os.path.realpath(link)
                    if sys.platform == 'win32':
                        rpath = rpath.replace('\\', '/')

                except:
                    return messages

                lstat = None
                if os.path.exists(rpath): 
                   lstat = sarracenia.stat(rpath)

                messages.extend(self.post1file(rpath, lstat))

        # path deleted

        elif lstat == None:
            messages.extend(self.post_delete(path,key=None,value=None,is_directory=is_directory))

        # path is a file

        elif os.path.isfile(path) or os.path.isdir(path):
            messages.extend(self.post_file(path, lstat))

        return messages

    def post1move(self, src, dst):
        #logger.debug("post1move %s %s" % (src,dst) )

        self.move_dir_lst = []

        messages = self.post_move(src, dst)

        for tup in self.move_dir_lst:
            src, dst = tup
            #logger.debug("deleting moved directory %s" % src )
            messages.extend(self.post_delete(src, 'newname', dst))

        return messages

    def process_event(self, event, src, dst):
        """
          return a list of messages.
        """
        #logger.debug("process_event %s %s %s " % (event,src,dst) )

        # delete

        if event == 'delete' :
            if event in self.o.fileEvents:
                return self.post1file(src, None)
            return []

        if event == 'rmdir' :
            if event in self.o.fileEvents:
                return self.post1file(src, None, is_directory=True)
            return []

        # move

        if event == 'move':
            if self.o.create_modify:
                return self.post1move(src, dst)

        # create or modify

        # directory : skipped, its content is watched
        #if self.o.recursive and os.path.isdir(src):
        #    dirs = list(map(lambda x: x[1][1], self.inl.items()))
        #    #logger.debug("skipping directory %s list: %s" % (src, dirs))

        # link ( os.path.exists = false, lstat = None )

        if os.path.islink(src):
            if 'link' in self.o.fileEvents:
                return self.post1file(src, None)
            return []

        # file : must exists
        #       (may have been deleted since event caught)

        if not os.path.exists(src): return []

        # file : must be old enough

        lstat = sarracenia.stat(src)

        if path_inflight_tooNew(self.o.inflight, lstat): return []

        # post it

        if event == 'mkdir' and 'mkdir' in self.o.fileEvents:
            return self.post1file(src, lstat, is_directory=True)
        elif self.o.create_modify: 
            return self.post1file(src, lstat)
        return []

    def set_blocksize(self, bssetting, fsiz):

        tfactor = 50 * 1024 * 1024

        if bssetting == 0:  ## default blocksize
            return tfactor

        elif bssetting == 1:  ## send file as one piece.
            return fsiz

        else:  ## partstr=i
            return bssetting

    def wakeup(self):
        #logger.debug("wakeup")

        # FIXME: Tiny potential for events to be dropped during copy.
        #     these lists might need to be replaced with watchdog event queues.
        #     left for later work. PS-20170105
        #     more details: https://github.com/gorakhargosh/watchdog/issues/392

        # pile up left events to process

        self.left_events.update(self.new_events)
        self.new_events = OrderedDict()

        # work with a copy events and keep done events (to delete them)

        self.cur_events = OrderedDict()
        self.cur_events.update(self.left_events)

        # loop on all events

        messages = []
        for key in self.cur_events:
            event, src, dst = self.cur_events[key]
            try:
                messages.extend(self.process_event(event, src, dst))
            except OSError as err:
                logger.error("skipping event that could not be processed: ({}): {}".format(
                    event, err))
                logger.debug("Exception details:", exc_info=True)
            self.left_events.pop(key)
        return messages

    def walk(self, src):
        """
          walk directory tree returning 1 message for each file in it.
        """
        logger.debug("walk %s" % src)

        # how to proceed with symlink

        if os.path.islink(src) and self.o.realpathPost:
            src = os.path.realpath(src)
            if sys.platform == 'win32':
                src = src.replace('\\', '/')

        messages = []

        # need to post root of tree first, so mode bits get propagated on creation.
        if src == self.o.post_baseDir :
            logger.debug("skip posting of post_baseDir {src}")
        else:
            messages.extend(self.post1file(src, sarracenia.stat(src), is_directory=True))

        # walk src directory, this walk is depth first... there could be a lot of time
        # between *listdir* run, and when a file is visited, if there are subdirectories before you get there.
        # hence the existence check after listdir (crashed in flow_tests of > 20,000)

        if self.o.recursive:
            for x in os.listdir(src):
                path = src + '/' + x
                # add path created
                if os.path.isdir(path):
                    messages.extend(self.walk(path))
                    continue

                if os.path.exists(path):
                    messages.extend(self.post1file(path, sarracenia.stat(path)))


        return messages

    def walk_priming(self, p):
        """
         Find all the subdirectories of the given path, start watches on them. 
         deal with symbolically linked directories correctly
        """
        if os.path.islink(p):
            realp = os.path.realpath(p)
            if sys.platform == 'win32':
                realp = realp.replace('\\', '/')

            logger.info("sr_watch %s is a link to directory %s" % (p, realp))
            if self.o.realpathPost:
                d = realp
            else:
                d = p + '/' + '.'
        else:
            d = p

        try:
            fs = sarracenia.stat(d)
            dir_dev_id = '%s,%s' % (fs.st_dev, fs.st_ino)
            if dir_dev_id in self.inl:
                return True
        except OSError as err:
            logger.warning("could not stat file ({}): {}".format(d, err))
            logger.debug("Exception details:", exc_info=True)

        if os.access(d, os.R_OK | os.X_OK):
            try:
                ow = self.observer.schedule(self.watch_handler,
                                            d,
                                            recursive=True)
                self.obs_watched.append(ow)
                self.inl[dir_dev_id] = (ow, d)
                logger.info(
                    "sr_watch priming watch (instance=%d) scheduled for: %s " %
                    (len(self.obs_watched), d))
            except:
                logger.warning("sr_watch priming watch: %s failed, deferred." %
                               d)
                logger.debug('Exception details:', exc_info=True)

                # add path created
                self.on_add('create', p, None)
                return True

        else:
            logger.warning(
                "sr_watch could not schedule priming watch of: %s (EPERM) deferred."
                % d)
            logger.debug('Exception details:', exc_info=True)

            # add path created
            self.on_add('create', p, None)
            return True

        return True

    def watch_dir(self, sld):
        logger.debug("watch_dir %s" % sld)

        if not features['watch']['present']:
            logger.critical("sr_watch needs the python watchdog library to be installed.")
            return []

        if self.o.force_polling:
            logger.info(
                "sr_watch polling observer overriding default (slower but more reliable.)"
            )
            self.observer = PollingObserver()
        else:
            logger.info(
                "sr_watch optimal observer for platform selected (best when it works)."
            )
            self.observer = Observer()

        self.obs_watched = []

        self.watch_handler = SimpleEventHandler(self)
        self.walk_priming(sld)

        logger.info(
            "sr_watch priming walk done, but not yet active. Starting...")
        self.observer.start()
        logger.info("sr_watch now active on %s posting to exchange: %s" %
                    (sld, self.o.post_exchange))

        if self.o.post_on_start:
            return self.walk(sld)
        else:
            return []

    def on_start(self):
        self.queued_messages = []
        self.primed = False

    def gather(self):
        """
           from sr_post.py/run 

           FIXME: really bad performance with large trees: It scans an entire tree
           before emitting any messages. Need to re-factor with iterator style so produce
           result in batch sized chunks incrementally.
        """
        #logger.debug("%s run partflg=%s, sum=%s, nodupe_ttl=%s basis=%s pbd=%s" % \
        #      ( self.o.component, self.o.partflg, self.o.sumflg, self.o.nodupe_ttl,
        #        self.o.nodupe_basis, self.o.post_baseDir ))
        #logger.debug("%s realpathPost=%s follow_links=%s force_polling=%s batch=%s"  % \
        #      ( self.o.component, self.o.realpathPost, self.o.follow_symlinks, \
        #        self.o.force_polling, self.o.batch ) )
        #logger.info("%s len(self.queued_messages)=%d" % \
        #     ( self.o.component, len(self.queued_messages) ) )

        pbd = self.o.post_baseDir

        if len(self.queued_messages) > self.o.batch:
            messages = self.queued_messages[0:self.o.batch]
            self.queued_messages = self.queued_messages[self.o.batch:]
            return messages

        elif len(self.queued_messages) > 0:
            messages = self.queued_messages
            self.queued_messages = []

            if self.o.sleep < 0:
                return messages
        else:
            messages = []

        if self.primed:
            return self.wakeup()

        cwd = os.getcwd()

        for d in self.o.postpath:

            # convert relative path to absolute.
            if d[0] != os.sep: d = cwd + os.sep + d

            d=self.o.variableExpansion(d)
            logger.debug("postpath = %s" % d)

            if self.o.sleep > 0:
                if features['watch']['present']:
                    messages.extend(self.watch_dir(d))
                else:
                    logger.critical("python watchdog package missing! Cannot watch directory")
                continue

            if os.path.isdir(d):
                logger.debug("postpath = %s" % d)
                messages.extend(self.walk(d))
            elif os.path.islink(d):
                messages.extend(self.post1file(d, None))
            elif os.path.isfile(d):
                messages.extend(self.post1file(d, sarracenia.stat(d)))
            else:
                logger.error("could not post %s (exists %s)" %
                             (d, os.path.exists(d)))

        if len(messages) > self.o.batch:
            self.queued_messages = messages[self.o.batch:]
            logger.info("len(queued_messages)=%d" % len(self.queued_messages))
            messages = messages[0:self.o.batch]

        self.primed = True
        return messages



#!/usr/bin/env python3

# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2020
#

import sys
import os
import os.path
import inspect
from sarracenia.config import Config

# this stub is included in v3 to allow v02 users to continue to use the familiar
# invocation of individual components, however, all set up is now done through
# central sr.py script.

sys.argv[-1] = os.path.basename(sys.argv[0]).replace('sr_', '').replace(
    '.py', '') + os.sep + sys.argv[-1]
args = [
    sys.executable,
    os.path.dirname(inspect.getfile(Config)) + os.sep + 'sr.py'
]
args.extend(sys.argv[1:])

os.execvp(sys.executable, args)



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2020
#

import logging

import sarracenia.moth
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class Message(FlowCB):
    """
       post messages to sarracenia.moth message queuing protocol destination.
    """
    def __init__(self, options):

        super().__init__(options,logger)

        if hasattr(self.o, 'post_broker'):
            props = sarracenia.moth.default_options
            props.update(self.o.dictify())

            if hasattr(self.o, 'topic' ):
                del self.o['topic']

            # adjust settings post_xxx to be xxx, as Moth does not use post_ ones.
            for k in [ 'broker', 'exchange', 'topicPrefix', 'exchangeSplit', 'topic' ]:
                post_one='post_'+k
                if hasattr( self.o, post_one ): 
                    #props.update({ k: getattr(self.o,post_one) } )
                    props[ k ] = getattr(self.o,post_one)

            self.poster = sarracenia.moth.Moth.pubFactory(props)

    def post(self, worklist):

        still_ok = []
        all_good=True
        for m in worklist.ok:
            if all_good and hasattr(self.poster,'putNewMessage') and self.poster.putNewMessage(m):
                still_ok.append(m)
            else:
                all_good=False
                worklist.failed.append(m)
        worklist.ok = still_ok

    def metricsReport(self) -> dict:
        if hasattr(self,'poster') and self.poster:
            return self.poster.metricsReport()
        else:
            return {}

    def on_housekeeping(self):
        if hasattr(self,'poster') and self.poster:
            m = self.poster.metricsReport()
            logger.info(
                f"messages: good: {m['txGoodCount']} bad: {m['txBadCount']} bytes: {m['txByteCount']}"
            )
            self.poster.metricsReset()
        else:
            logger.info( "no metrics available" )

    def on_start(self):
        if hasattr(self,'poster') and self.poster:
            self.poster.putSetup()
        logger.info('starting')

    def on_stop(self):
        if hasattr(self,'poster') and self.poster:
            self.poster.close()
        logger.info('closing')



"""
Plugin longflow.py:
    This plugin is strictly for self-test purposes.
    Creates a header 'toolong' that is longer than 255 characters, and so gets truncated.
    Each header in a message that is too long should generate a warning message in the sarra log.
    flow_test checks for the 'truncated' error message.
    Put some utf characters in there to make it interesting... (truncation complex.)

Usage:
    flowcb sarracenia.flowcb.accept.longflow.LongFlow
"""

import logging
import os, stat, time
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class LongFlow(FlowCB):
    def __init__(self, options):
        super().__init__(options, logger)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            logger.info('setting toolong message key')
            message['toolong'] = '1234567890ßñç' * 26



"""
    Example of posting data to a Sarracenia Data Pump.

"""
import sarracenia.moth
import sarracenia
import sarracenia.credentials
from sarracenia.config import default_config

import os
import time
import socket
import sys

"""
  supply broker argument as a command line argument.

"""
if len(sys.argv) > 1:
    broker = sys.argv[1]
else:
    broker = 'amqp://tfeed:HungryCat@localhost'

cfg = default_config()
#cfg.logLevel = 'debug'
cfg.broker = sarracenia.credentials.Credential( broker )
cfg.exchange = 'xsarra'
cfg.post_baseUrl = 'http://host'
cfg.post_baseDir = '/tmp'
cfg.topicPrefix = [ 'v03', 'post' ]
cfg.logLevel = 'debug'

print('cfg: %s' % cfg)

# moth wants a dict as options, rather than sarracenia.config.Config instance.
posting_engine = sarracenia.moth.Moth.pubFactory(cfg.dictify())

sample_fileName = '/tmp/sample.txt'
sample_file = open(sample_fileName, 'w')
sample_file.write("""
CACN00 CWAO 161800
PMN
160,2021,228,1800,1065,100,-6999,20.49,43.63,16.87,16.64,323.5,9.32,27.31,1740,317.8,19.22,1.609,230.7,230.7,230.7,230.7,0,0,0,16.38,15.59,305.
9,17.8,16.38,19.35,55.66,15.23,14.59,304,16.67,3.844,20.51,18.16,0,0,-6999,-6999,-6999,-6999,-6999,-6999,-6999,-6999,0,0,0,0,0,0,0,0,0,0,0,0,0,
13.41,13.85,27.07,3473
""")
sample_file.close()

# you can supply msg_init with your files, it will build a message appropriate for it.
m = sarracenia.Message.fromFileData(sample_fileName, cfg, sarracenia.stat(sample_fileName))

# here is the resulting message.
print(m)

# feed the message to the posting engine.
posting_engine.putNewMessage(m)

# when done, should close... cleaner...
posting_engine.close()



""" 
   msg_pclean_f90 module: file propagation test for Sarracenia components (in flow test)
   https://github.com/MetPX/sr_insects/

"""

from difflib import Differ
import filecmp
import logging
import os
import random

from sarracenia.flowcb.pclean import PClean
from sarracenia import timestr2flt, nowflt

logger = logging.getLogger(__name__)


class PClean_F90(PClean):
    """ 
     functionality within the flow_tests of the sr_insects project.
     This plugin class receive a msg from xflow_public and check propagation of the underlying file

     - it checks if the propagation was ok
     - it randomly set a new test file with a different type in the watch dir (f31 amqp)
     - it posts the product to be treated by f92
     - when the msg for the extension file comes back, recheck the propagation

    When a product is not fully propagated, the error is reported and the test is considered as a
    failure. It also checks if the file differs from original
    """
    def after_accept(self, worklist):

        logger.info("start len(worklist.incoming) = %d" %
                    len(worklist.incoming))

        outgoing = []

        for msg in worklist.incoming:

            result = True
            f20_path = '/' + msg['relPath'].replace(
                "{}/".format(self.all_fxx_dirs[1]), self.all_fxx_dirs[0])
            path_dict = self.build_path_dict(self.all_fxx_dirs[2:],
                                             '/' + msg['relPath'])
            ext = self.get_extension('/' + msg['relPath'])
            logger.info('looking at: %s' % msg['relPath'])
            logger.info('path_dict: %s' % path_dict)

            for fxx_dir, path in path_dict.items():
                # f90 test
                logger.info('for looping: %s' % path)
                if not (os.path.isfile(path) or os.path.islink(path)):
                    # propagation check to all path except f20 which is the origin
                    err_msg = "file not in folder {} with {:.3f}s elapsed"
                    lag = nowflt() - timestr2flt(msg['pubTime'])
                    logger.error(err_msg.format(fxx_dir, lag))
                    logger.debug("file missing={}".format(path))
                    result = False
                    worklist.failed.append(msg)
                    break
                elif ext not in self.test_extension_list and not filecmp.cmp(
                        f20_path, path):
                    # file differ check: f20 against others
                    logger.error(
                        "skipping, file differs from f20 file: {}".format(
                            path))
                    with open(f20_path, 'r', encoding='iso-8859-1') as f:
                        f20_lines = f.readlines()
                    with open(path, 'r', encoding='iso-8859-1') as f:
                        f_lines = f.readlines()
                    diff = Differ().compare(f20_lines, f_lines)
                    diff = [d for d in diff
                            if d[0] != ' ']  # Diffs without context
                    logger.info("a: len(%s) = %d" % (f20_path, len(f20_lines)))
                    logger.info("b: len(%s) = %d" % (path, len(f_lines)))
                    logger.info("diffs found:\n{}".format("".join(diff)))

            if not result:
                logger.info('queued for retry because propagation not done yet.')
                continue

            if ext not in self.test_extension_list:
                # prepare next f90 test
                test_extension = self.test_extension_list[self.ext_count % len(
                    self.test_extension_list)]
                self.ext_count += 1
                # pick one test identified by file extension
                src = '/' + msg['relPath']  # src file is in f30 dir
                dest = "{}{}".format(
                    src, test_extension
                )  # format input file for extension test (next f90)

                try:
                    if test_extension == '.slink':
                        os.symlink(src, dest)
                        logger.info('symlinked %s %s' % (src, dest))
                    elif test_extension == '.hlink':
                        os.link(src, dest)
                        logger.info('hlinked %s %s' % (src, dest))
                    elif test_extension == '.moved':
                        os.rename(src, dest)
                        logger.info('moved %s %s' % (src, dest))
                    else:
                        logger.error("test '{}' is not supported".format(
                            test_extension))
                except FileNotFoundError as err:
                    # src is not there
                    logger.error("test failed: {}".format(err))
                    logger.debug("Exception details:", exc_info=True)
                    result = False
                except FileExistsError as err:
                    # dest is already there
                    logger.error(
                        'skipping, found a moving target {}'.format(err))
                    logger.debug("Exception details:", exc_info=True)
                    result = False
            else:
                logger.info('ext not in test_extenion_list')

            if 'toolong' in msg:
                # cleanup
                del msg['toolong']

            if result:
                outgoing.append(msg)
            else:
                worklist.rejected.append(msg)

        worklist.incoming = outgoing
        logger.info("end len(worklist.incoming) = %d" % len(worklist.incoming))
        logger.info("end len(worklist.rejected) = %d" % len(worklist.rejected))




#
# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Shared Services Canada, 2020
#

import copy
import importlib
import logging
import sys

entry_points = [

    'ack', 'after_accept', 'after_post', 'after_work', 'destfn', 'do_poll', 
    'download', 'gather', 'metricsReport', 'on_cleanup', 'on_declare', 'on_features',
    'on_housekeeping', 'on_sanity', 'on_start', 'on_stop', 
    'please_stop', 'poll', 'post', 'report', 'send', 

]

logger = logging.getLogger(__name__)


class FlowCB:
    """
    Flow Callback is the main class for implementing plugin customization to flows.

    sample activation in a configuration file:
    
    flowCallback sarracenia.flowcb.name.Name
    
    will instantiate an object of that type whose appropriately name methods
    will be called at the right time.
    
    __init__ accepts options as an argument.
    
    options is a sarracenia.config.Config object, used to override default behaviour
    
    a setting is declared in a configuration file like so::
    
        set sarracenia.flowcb.filter.log.Log.level debug
    
    (the prefix for the setting matches the type hierarchy in flowCallback)
    the plugin should get the setting::
    
        options.level = 'debug'
    
    
    worklist given to on_plugins...
    
    * worklist.incoming --> new messages to continue processing
    * worklist.ok       --> successfully processed
    * worklist.rejected --> messages to not be further processed.
    * worklist.failed   --> messages for which processing failed. Failed messages will be retried.
    * worklist.directories_ok --> list of directories created during processing.
    
    Initially, all messages are placed in incoming.
    if a plugin entry_point decides:
    
    - a message is not relevant, it is moved to the rejected worklist. 
    - all processing has been done, it moves it to the ok worklist
    - an operation failed and it should be retried later, append it to the failed
      worklist
    
    Do not remove any message from all lists, only move messages between them.
    it is necessary to put rejected messages in the appropriate worklist
    so they can be acknowledged as received. Messages can only removed after ack.
    
    
    def __init__(self,options) -> None::

        Task: initialization of the flowCallback at instantiation time.

        usually contains:

        self.o = options

    def ack(self,messagelist) -> None::

        Task: acknowledge messages from a gather source.

    def gather(self) -> list::

        Task: gather messages from a source... return a list of messages.

              in a poll, gather is always called, regardless of vip posession.
              in all other components, gather is only called when in posession
              of the vip.
        return []

    def after_accept(self,worklist) -> None::

         Task: just after messages go through accept/reject masks,
               operate on worklist.incoming to help decide which messages to process further.
               and move messages to worklist.rejected to prevent further processing.
               do not delete any messages, only move between worklists.

    def after_work(self,worklist) -> None::

        Task: operate on worklist.ok (files which have arrived.)

        All messages on the worklist.ok list have been acknowledged, so to suppress posting
        of them, or futher processing, the messages must be removed from worklist.ok.

        worklist.failed processing should occur in here as it will be zeroed out after this step.
        The flowcb/retry.py plugin, for example, processes failed messages.

    def destfn(self,msg) -> str::

         Task: look at the fields in the message, and perhaps settings and
               return a new file name for the target of the send or download.

         kind of a last resort function, exists mostly for sundew compatibility.
         can be used for selective renaming using accept clauses.

    def download(self,msg) -> bool::

         Task: looking at msg['new_dir'], msg['new_file'], msg['new_inflight_file'] 
               and the self.o options perform a download of a single file.
               return True on a successful transfer, False otherwise.

               if self.o.dry_run is set, simulate the output of a download without
               performing it.

         This replaces built-in download functionality, providing an override.
         for individual file transfers. ideally you set checksums as you download.
            
    def metricsReport(self) -> dict:

        Return a dictionary of metrics. Example: number of messages remaining in retry queues.

    def on_cleanup(self) -> None::
        allow plugins to perform additional work after broker resources are eliminated.
        local state files are still present when this runs.

    def on_declare(self) -> None::
        local state files are still already present when this runs.
        allow plugins to perform additional work besides broker resource setup.

    def on_housekeeping(self) -> None::

         do periodic processing.

    def on_start(self) -> None::

         After the connection is established with the broker and things are instantiated, but
         before any message transfer occurs.

    def on_stop(self) -> None::
        
         what it says on the tin... clean up processing when stopping.         

    def poll(self) -> list::

        Task: gather messages from a destination... return a list of messages.
              works like a gather, but...

              When specified, poll replaces the built-in poll of the poll component.
              it runs only when the machine running the poll has the vip.
              in components other than poll, poll is never called.
        return []

    def post(self,worklist) -> None::
         
         Task: operate on worklist.ok, and worklist.failed. modifies them appropriately.
               message acknowledgement has already occurred before they are called.

         to indicate failure to process a message, append to worklist.failed.
         worklist.failed processing should occur in here as it will be zeroed out after this step.

    def send(self,msg) -> bool::

         Task: looking at msg['new_dir'], msg['new_file'], and the self.o options perform a transfer
               of a single file.
               return True on a successful transfer, False otherwise.

               if self.o.dry_run is set, simulate the output of a send without
               performing it.

         This replaces built-in send functionality for individual files.

    def please_stop(self):
         Pre-warn a flowcb that a stop has been requested, allowing processing to wrap up
         before the full stop happens.

    """
    def __init__(self, options, class_logger=None):
        self.o = options
        self.stop_requested = False

        if hasattr(self.o,'logFormat'):
            logging.basicConfig(format=self.o.logFormat,
                            level=getattr(logging, self.o.logLevel.upper()))

        if hasattr(self.o,'logLevel') and class_logger:
            class_logger.setLevel(getattr(logging, self.o.logLevel.upper()))


    def please_stop(self):
        """
           flow callbacks should not time.sleep for long periods, but only nap and check
           between naps if a stop has been requested.
        """
        self.stop_requested = True

def load_library(factory_path, options):
    """
       Loading the entry points for a python module. It searches 
       the normal python module path using the importlib module. 

       the factory_path is a combined file specification with a dot separator
       with a special last entry being the name of the class within the file.

       factory_path  a.b.c.C

       means import the module named a.b.c and instantiate an object of type
       C. In that class-C object, look for the known callback entry points. 

       or C might be guessed by the last class in the path not following
       python convention by not starting with a capital letter, in which case,
       it will just guess.

       re
       note that the ~/.config/sr3/plugins will also be in the python library 
       path, so modules placed there will be found, in addition to those in the
       package itself in the *sarracenia/flowcb*  directory

       callback foo  -> foo.Foo
                        sarracenia.flowcb.foo.Foo

       callback foo.bar -> foo.bar.Bar
                           sarracenia.flowcb.foo.bar.Bar
                           foo.bar
                           sarracenia.flowcb.foo.bar 
    """

    if not '.' in factory_path:
        packagename = factory_path
        classname =factory_path.capitalize()
    else:
        if factory_path.split('.')[-1][0].islower():
            packagename = factory_path
            classname = factory_path.split('.')[-1].capitalize()
        else:
            packagename, classname = factory_path.rsplit('.', 1)

    try:
        module = importlib.import_module('sarracenia.flowcb.' + packagename)
        class_ = getattr(module, classname)
    except ModuleNotFoundError:
        module = importlib.import_module(packagename)
        class_ = getattr(module, classname)
 
    if hasattr(options, 'settings'):
        opt = copy.deepcopy(options)
        # strip off the class prefix.
        if factory_path in options.settings:
            for s in options.settings[factory_path]:
                setattr(opt, s, options.settings[factory_path][s])
    else:
        opt = options

    plugin = class_(opt)
    return plugin




"""

Retrieves file keys from an Amazon AWS S3 bucket.

callback poll.s3bucket

pollUrl https://s3-us-west-1.amazonaws.com//files.airnowtech.org/airnow/today/HourlyData_${YYYYMMDD-70m}

( https://metpx.github.io/sarracenia/Reference/sr3_options.7.html#variables ) for details on variable substitutions )

to poll AWS for a particular site's level 2 nexrad data:

#check for the RKJK radar site
declare env Site RKJK

#for readar products from the current date as of seven minutes ago.
varTimeOffset -7m
pollUrl https://s3-us-east-1.amazonaws.com//nexrad-level-2/${YYYY}/${MM}/${DD}/${Site}

#check every five minutes.
sleep 300


"""
import boto3
from botocore import UNSIGNED
from botocore.client import Config
import datetime
import logging
import paramiko
import sarracenia
from sarracenia.flowcb import FlowCB
import sys
import urllib.parse
import urllib.request

logger = logging.getLogger(__name__)


class S3bucket(FlowCB):
    def __init__(self, options):

        super().__init__(options,logger)

        self.minutetracker = datetime.datetime.utcnow() + datetime.timedelta(minutes=-70)
        
        logger.info( f" url: {self.o.pollUrl} " )
        ppu = urllib.parse.urlparse(self.o.pollUrl)

        logger.info( f" host: {ppu.netloc} path:{ppu.path} " )
        self.service = ppu.netloc.split('-')[0]
        self.region = ppu.netloc.split('.')[0][3:]
        self.bucket = ppu.path.split('/')[2]
        self.prefix = '/'.join(ppu.path.split('/')[3:])
        logger.info( f" service: {self.service} region: {self.region} bucket:{self.bucket}  prefix={self.prefix}" )

    def poll(self):

        last_hour_date_time = datetime.datetime.now() - datetime.timedelta(hours=1)
        keys = []
        keysizes = []

        current_prefix = self.o.variableExpansion( self.prefix )
        logger.info("current_prefix: {current_prefix}" )

        # Gets the list of active weather station ICAOs from
        s3 = boto3.client( self.service,
            region_name=self.region, config=Config(signature_version=UNSIGNED, retries={ 'max_attempts': 3 }))
        try:
            for obj in s3.list_objects(Bucket=self.bucket, Prefix=current_prefix)['Contents']:
                keys.append(obj['Key'])
                keysizes.append(obj['Size'])
        except KeyError as e:
            print(e)

        i = 0
        gathered_messages = []
        while i < len(keys):
            logger.info("found {}".format(keys[i]))

            fakestat = paramiko.SFTPAttributes()
            fakestat.st_size = keysizes[i]

            m = sarracenia.Message.fromFileInfo(keys[i].replace(self.prefix,'',1), self.o, fakestat)
            gathered_messages.append(m)

            i += 1
        return gathered_messages



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2020
#

import logging

from sarracenia import naturalSize
import sarracenia.moth
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class Message(FlowCB):
    """
       gather messages from a sarracenia.moth message queuing protocol source.
    """
    def __init__(self, options) -> None:

        super().__init__(options,logger)

        self.od = sarracenia.moth.default_options
        self.od.update(self.o.dictify())

        if hasattr(self.o, 'broker') and self.o.broker:
            self.consumer = sarracenia.moth.Moth.subFactory(self.od)
        else:
            logger.critical('missing required broker specification')

    def gather(self) -> list:
        """
           return a current list of messages.
        """
        if hasattr(self,'consumer') and hasattr(self.consumer,'newMessages'):
            return self.consumer.newMessages()
        else:
            logger.warning( f'not connected. Trying to connect to {self.o.broker}')
            self.consumer = sarracenia.moth.Moth.subFactory(self.od)
            return []

    def ack(self, mlist) -> None:

        if not hasattr(self,'consumer'):
            return

        for m in mlist:
            # messages being re-downloaded should not be re-acked, but they won't have an ack_id (see issue #466)
            self.consumer.ack(m)

    def metricsReport(self) -> dict:
        if hasattr(self,'consumer') and hasattr(self.consumer,'metricsReport'):
           return self.consumer.metricsReport()
        else:
           return {}

    def on_housekeeping(self) -> None:

        if not hasattr(self,'consumer'):
            return

        if hasattr(self.consumer, 'metricsReport'):
            m = self.consumer.metricsReport()
            average = (m['rxByteCount'] /
                   m['rxGoodCount'] if m['rxGoodCount'] != 0 else 0)
            logger.info( f"messages: good: {m['rxGoodCount']} bad: {m['rxBadCount']} " +\
               f"bytes: {naturalSize(m['rxByteCount'])} " +\
               f"average: {naturalSize(average)}" )
            self.consumer.metricsReset()

    def on_stop(self) -> None:
        if hasattr(self,'consumer') and hasattr(self.consumer, 'close'):
            self.consumer.close()
        logger.info('closing')



#/usr/bin/env python3
"""

replacement for tail -f to use on Windows systems that don't have it.

started with: https://gist.github.com/amitsaha/5990310

"""

#!/usr/bin/python3

import os, sys, time


def tail_file(filename, nlines):
    with open(filename) as qfile:
        qfile.seek(0, os.SEEK_END)
        endf = position = qfile.tell()
        linecnt = 0
        while position >= 0:
            qfile.seek(position)
            next_char = qfile.read(1)
            if next_char == "\n" and position != endf - 1:
                linecnt += 1

            if linecnt == nlines:
                break
            position -= 1

        if position < 0:
            qfile.seek(0)

        print(qfile.read(), end='')

        while True:
            l = qfile.readline()
            if (l):
                sys.stdout.write(l)
            else:
                time.sleep(0.2)


def main():
    filename = sys.argv[1]
    #nlines = int(sys.argv[2])
    tail_file(filename, 10)


if __name__ == '__main__':
    main()



"""

  a destfn plugin script is used by senders or subscribers to do complex file naming.
  this is an API demonstrator that prefixes the name delivered with 'renamed\_'::

     filename DESTFNSCRIPT=sarracenia.flowcb.destfn.sample.Sample

  An alternative method of invocation is to apply it selectively::

     accept k* DESTFNSCRIPT=sarracenia.flowcb.destfn.sample.Sample

  As with other flowcb plugins, the import will be done using normal
  python import mechanism equivalent to::

     import sarracenia.flowcb.destfn.sample

  and then in that class file, there is a Sample class, the sample class
  contains the destfn method, or entry_point.
 
  The destfn routine consults the fields in the given message, and based on them,
  return a new file name for the file to have after transfer (download or send.)

  the routines have access to the settings via options provided to init,
  accessed, by convention, as self.o.

  The routine can also modify fields create new ones in the message.

  the destfn routine returns the new name of the file.

"""

import logging

from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)

class Sample(FlowCB):

      def destfn(self,msg) -> str:

          logger.info('before: m=%s' % msg )
          relPath = msg["relPath"].split('/')
          msg['destfn_added_prefix'] = 'renamed_'
          return 'renamed_' + relPath[-1]




#!/usr/bin/env python3

# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2020
#

import sys
import os
import inspect
import sarracenia.config
import sarracenia.flow
import sarracenia.flow.post
import logging


def main():
    logger = logging.getLogger()
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s %(funcName)s %(message)s',
        level=logging.DEBUG)

    logger.setLevel(logging.INFO)

    if (len(sys.argv) > 1) and (sys.argv[1] == '--debug'):
        logger.setLevel(logging.DEBUG)

    cfg1 = sarracenia.config.default_config()

    cfg1.override({
        'component': 'post',
        'directory': os.getcwd(),
        'acceptUnmatched': True,
        'action': 'foreground'
    })

    cfg1.override(sarracenia.flow.post.default_options)

    cfg1.parse_args(isPost=True)

    if hasattr(cfg1, 'config'):
        cfg2 = sarracenia.config.one_config('post', cfg1.config, isPost=True)
    else:
        logger.critical('no posting configuration specified')
        return

    if hasattr(cfg2, 'component'):
        post_flow = sarracenia.flow.Flow.factory(cfg2)
        post_flow.run()


if __name__ == "__main__":
    main()



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, 2008-2020
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
#

import logging

import sarracenia
from sarracenia.moth.amqp import AMQP, amqp_ss_maxlen, default_options

import time
import queue
import threading

logger = logging.getLogger(__name__)

class AMQPConsumer(AMQP):
    """
        Extension of the existing AMQP implementation, replacing basic_get with a consumer.
        https://github.com/celery/py-amqp#quick-overview

        Set amqp_consumer True in the config to use this instead of the regular AMQP class.

        TODO: 
          - how does this work with the batch and prefetch options? 
    """

    def __init__(self, props, is_subscriber) -> None:
        super().__init__(props, is_subscriber)
        self._raw_msg_q = queue.Queue() # Queue is thread safe
        self._consumer_thread = None
        self._thread_please_stop = False

        # control log level in config file:
        # set sarracenia.moth.amqpconsumer.AMQPConsumer.logLevel debug
        me = "%s.%s" % (__class__.__module__, __class__.__name__)
        if ('settings' in self.o) and (me in self.o['settings']):
            for s in self.o['settings'][me]:
                self.o[s] = self.o['settings'][me][s]
            if 'logLevel' in self.o['settings'][me]:
                logger.setLevel(self.o['logLevel'].upper())

    def __consumer_setup(self) -> None:
        """ Start consuming from the queue.
        """
        if not self.connection:
            self.getSetup()
        # docs.celeryq.dev/projects/amqp/en/latest/reference/amqp.channel.html#amqp.channel.Channel.basic_consume
        self.channel.basic_consume(self.o['queueName'], no_ack=False, callback=self.__get_on_message)
        self.__stop_thread() # make sure it's not already running
        self._consumer_thread = threading.Thread(target=self.__drain_events)
        self._consumer_thread.start()

    def __drain_events(self):
        """ Calls drain_events on the connection until told to stop.
        """
        logger.debug("thread starting")
        while not self._thread_please_stop:
            # This blocks until there's an event to deal with
            try:
                self.connection.drain_events(timeout=2) # TODO configurable timeout?
            except TimeoutError:
                pass
            except Exception as e:
                logger.error(f"exception occurred: {e}")
                logger.debug('Exception details: ', exc_info=True)
        logger.debug("thread stopping")
        self._consumer_thread = None

    def __get_on_message(self, msg):
        """ Callback for AMQP basic_consume, called when the broker sends a new message.
        """
        logger.debug(f"new message pushed from broker: {msg.body}")
        # This will block until the msg can be put in the queue
        self._raw_msg_q.put(msg)

    def __stop_thread(self):
        # need to stop consuming - tell the thread to stop, then join it and wait
        self._thread_please_stop = True
        if self._consumer_thread:
            self._consumer_thread.join()
        self._consumer_thread = None
        self._thread_please_stop = False # if True, the thread won't start again

    def _amqp_setup_signal_handler(self, signum, stack):
        logger.info("ok, asked to stop")
        self.please_stop=True
        self.__stop_thread()

    def getCleanUp(self) -> None:
        self.__stop_thread()
        super().getCleanUp()

    def getNewMessage(self) -> sarracenia.Message:
        """ Mostly a copy of moth.amqp.AMQP's getNewMessage.
        """

        if not self.is_subscriber:  #build_consumer
            logger.error("getting from a publisher")
            return None

        try:
            if not self.connection:
                self.getSetup()

            if not self._consumer_thread:
                self.__consumer_setup()

            try:
                # don't block waiting for the queue to be available, better to just try again later
                raw_msg = self._raw_msg_q.get_nowait()
            except queue.Empty:
                raw_msg = None
            
            if (raw_msg is None) and (self.connection.connected):
                return None
            else:
                self.metrics['rxByteCount'] += len(raw_msg.body)
                try: 
                    msg = self._msgRawToDict(raw_msg)
                except Exception as err:
                    logger.error("message decode failed. raw message: %s" % raw_msg.body )
                    logger.debug('Exception details: ', exc_info=True)
                    msg = None
                if msg is None:
                    self.metrics['rxBadCount'] += 1
                    return None
                else:
                    self.metrics['rxGoodCount'] += 1
                if hasattr(self.o, 'fixed_headers'):
                    for k in self.o.fixed_headers:
                        msg[k] = self.o.fixed_headers[k]
                logger.debug("new msg: %s" % msg)
                return msg
        except Exception as err:
            logger.warning("failed %s: %s" %
                           (self.o['queueName'], err))
            logger.debug('Exception details: ', exc_info=True)

        if not self.o['message_strategy']['stubborn']:
            return None

        logger.warning('lost connection to broker')
        self.close()
        time.sleep(1)
        return None

    def close(self) -> None:
        try: 
            self.__stop_thread()
        except:
            pass
        super().close()



import logging
import os

from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class DeleteFlowFiles(FlowCB):
    """
       This is a custom callback class for the sr_insects flow tests.
       delete files for messages in two directories.
    """

    def after_accept(self, worklist):

        for m in worklist.incoming:

            f = "%s%s%s" % (m['new_dir'], os.sep, m['new_file'])
            try:
                os.unlink(f)
                os.unlink(f.replace('/cfr/', '/cfile/'))
                worklist.ok.append(m)
                logger.info("msg_delete: %s" % f)
            except OSError as err:
                logger.error("could not unlink {}: {}".format(f, err))
                logger.debug("Exception details:", exc_info=True)
                worklist.failed.append(m)
        worklist.incoming = []



"""
    sarracenia.flowcb.send.email.Email is an sr3 sender plugin. Once a file is 
    posted, the plugin matches the topic(what the filename begins with) to the
    file name and sends the appropriate emails.

    Usage:
      1. Need the following variables in an sr_sender config defined: file_email_to, file_email_relay
         Optionally, you can also provide a sender name/email as file_email_form:

            file_email_to AACN27 muhammad.taseer@canada.ca, test@test.com
            file_email_relay email.relay.server.ca
            file_email_from santa@canada.ca

      2. In the config file, include the following line:

            callback send.email

      3. sr_sender foreground emails.conf

    Original Author: Wahaj Taseer - June, 2019
"""

from email.message import EmailMessage
import logging
import os.path
import re
from sarracenia.flowcb import FlowCB
import smtplib

logger = logging.getLogger(__name__)


class Email(FlowCB):
    def __init__(self, options):

        super().__init__(options,logger)
        self.o.add_option('file_email_command', 'str', '/usr/bin/mail')
        self.o.add_option('file_email_to', 'list')
        self.o.add_option('file_email_from', 'str')
        self.o.add_option('file_email_relay', 'str')

    def send(self, msg):

        # have a list of email destinations...
        logger.debug("email: %s" % self.o.file_email_to)
        ipath = os.path.normpath(msg['relPath'])

        # loop over all the variables from config file, if files match, send via email
        for header in self.o.file_email_to:
            file_type, emails = header.split(' ', 1)
            emails = [x.strip(' ') for x in emails.split(',')]

            # check if the file arrived matches any email rules
            if re.search('^' + file_type + '.*', msg['new_file']):

                for recipient in emails:
                    logger.debug('sending file %s to %s' % (ipath, recipient))

                    with open(ipath) as fp:
                        emsg = EmailMessage()
                        emsg.set_content(fp.read())

                    try:
                        sender = self.o.file_email_from
                        if not sender:
                            sender = 'sarracenia-emailer'
                    except AttributeError:
                        sender = 'sarracenia-emailer'

                    logger.debug("Using sender email: " + sender)

                    emsg['Subject'] = msg['new_file']
                    emsg['From'] = sender
                    emsg['To'] = recipient

                    try:
                        email_relay = self.o.file_email_relay
                        if not email_relay:
                            raise AttributeError()
                    except AttributeError:
                        logger.error(
                            'file_email_relay config NOT defined, please define an SMTP (relay) server'
                        )

                    logger.debug("Using email relay server: " + email_relay)
                    s = smtplib.SMTP(email_relay)
                    s.send_message(emsg)
                    s.quit()

                    logger.info('sent file %s to %s' % (ipath, recipient))

        return True



"""
Posts updated files of airnowtech. 
Compatible with Python 3.5+.

usage:
	in an sr3 poll configuration file:

	pollUrl http://files.airnowtech.org/?prefix=airnow/today/

	callback airnow

STATUS: unknown... need some authentication, or perhaps the method has changed.
        does not seem to work out of the box.
"""

import datetime
import logging
import paramiko
import requests
import sarracenia
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class Airnow(FlowCB):

    def poll(self):

        sleep = self.o.sleep

        gathered_messages = []
        for Hours in range(1, 3):
            last_hour_date_time = datetime.datetime.now() - datetime.timedelta(
                hours=Hours)
            Filename = 'HourlyData_%s.dat' % last_hour_date_time.strftime(
                '%Y%m%d%H')
            logger.debug("poll_airnow_http Filename: %s" % Filename)
            URL = self.o.pollUrl + '/' + Filename
            logger.info('INFO %s ' % URL)
            #resp = requests.get(self.o.pollUrl + '/' + Filename)
            resp = requests.get(URL)
            if resp.ok:
                mtime = datetime.datetime.strptime(resp.headers['last-modified'],\
                     '%a, %d %b %Y %H:%M:%S %Z')
                last_poll = datetime.datetime.utcnow() + datetime.timedelta(
                    seconds=-sleep)
                logger.info(mtime)
                logger.info(last_poll)

                fakeStat = paramiko.SFTPAttributes()
                fakeStat.st_size = resp.headers['content-length']

                # convert datetime to numeric timestamp from beginning of POSIX epoch.
                fakeStat.st_mtime = mtime.timestamp()
                fakeStat.st_atime = mtime.timestamp()
                fakeStat.st_mode = 0o644

                m = sarracenia.Message.fromFileInfo(Filename, self.o, fakeStat)
                gathered_messages.append(m)

                logger.info('mtime: %s  last_pollL %s' % (mtime, last_poll))

        return gathered_messages



from sarracenia.moth import Moth

import logging

logger = logging.getLogger(__name__)


class AMQ1(Moth):
    def __init__(self, broker, props, is_subscriber):
        """
            AMQP 1.0 library to be built with libqpid-proton (the only free amqp 1.0 library around.)

            stub, not implemented
        """
        super().__init__(broker, props, is_subscriber)

        logger.error(
            "__init__ AMQP 1.0 Moth using qpid-proton library: not implemented"
        )



from sarracenia.identity import Identity

default_value = "None"


def set_default_value(value):
    global default_value
    default_value = value

class Arbitrary(Identity):
    """
      For applications where there is no known way of determining equivalence, allow them to supply
      an arbitrary tag, that can be used to compare products for duplicate suppression purposes.

      use setter to set the value... some sort of external checksum algorithm that cannot be reproduced.
     """
    def __init__(self):
        global default_value
        self._value = default_value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @staticmethod
    def registered_as():
        return 'a'

    def set_path(self, path):
        pass

    def update(self, chunk):
        pass



"""
Plugin save.py:
    Converts a consuming component into one that writes the queue into a file. 
    If there is some sort of problem with a component, then add callback save and restart.

    The messages will accumulate in a save file in ~/.cache/<component>/<config>/ ??<instance>.save
    When the situation is returned to normal (you want the component to process the data as normal):
    * remove the callback save
    * note the queue this configuration is using to read (should be in the log on startup.)
    * run an sr_shovel with -restore_to_queue and the queue name.

Options:
    save.py takes 'msgSaveFile' as an argument.
    "_9999.sav" will be added to msgSaveFile, where the 9999 represents the instance number.
    As instances run in parallel, rather than sychronizing access, just writes to one file per instance.

Usage:
    callback accept.save
    msgSaveFile x
"""

import jsonpickle
import logging
from sarracenia.flowcb import FlowCB

logger = logging.getLogger('__name__')


class Save(FlowCB):
    def __init__(self, options):
        super().__init__(options, logger)

        self.o.add_option('msgSaveFile', 'str')
        if not self.o.msgSaveFile:
            logger.error("msg_save: setting msgSaveFile setting is mandatory")
            return

        self.msgSaveFile = open(self.o.msgSaveFile, "a")
        logger.debug("msg_save initialized")

    def after_accept(self, worklist):
        for message in worklist.incoming:
            self.msgSaveFile.write(jsonpickle.encode(message) + '\n')
            self.msgSaveFile.flush()
            logger.info("msg_save saving msg with topic:%s (aborting further processing)" % message['topic'])



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, 2008-2022
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://metpx.github.io/sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
#

import collections
import copy
import json
import logging

from paho.mqtt.properties import Properties
from paho.mqtt.packettypes import PacketTypes
import paho.mqtt.client

import sarracenia
from sarracenia.postformat import PostFormat
from sarracenia.moth import Moth
import signal
import ssl
import threading
import time
from urllib.parse import unquote

logger = logging.getLogger(__name__)


default_options = {
    'manuel_ack': False,
    'batch': 25,
    'clean_session': False,
    'max_inflight_messages': 20000, # https://pypi.org/project/paho-mqtt/ ... says default is 20, raised because of data loss
    #'max_queued_messages' : 20000,
    'no': 0,
    'prefetch': 25,
    #'receiveMaximum': 65535, # https://docs.oasis-open.org/mqtt/mqtt/v5.0/os/mqtt-v5.0-os.html#_Receive_Maximum
    'topicPrefix': ['v03']
}

class MQTT(Moth):
    """
       Message Queue Telemetry Transport support.
           talks to an MQTT broker.  Tested with mosquitto. requires MQTTv5

       -  broker url schemes:  mqtt, mqtts, mqttw, mqttws

       Sarracenia Concept mapping from AMQP:

           AMQP -> MQTT topic hierarcy mapping: 
           exchange: xpublic,  topic prefix:   v03 subtopic:   /my/favourite/directory
           results in :   xpublic/v03/my/favourite/directory

           AMQP queue -> MQTT group-id (queue maps to queue, client-id should map to  queue+instance, I think.)

       deviation from standard: group-id lengths are not checked, neither are client-id's.
           Sarracenia routinely generates such ids with 40 to 50 characters in them.
           the MQTTv3.1 standard specifies a maximum length of 23 characters, as the minimum
           a compliant broker must support.  
           https://www.eclipse.org/lists/mosquitto-dev/msg00433.html#:~:text=It%20is%2065535%20bytes.,the%20limit%20in%20mqtt%20v3.
           Both mosquitto and EMQ supports 65535 chars for that field even in v3.1, so enforcing limit seems counter-productive.

       problems with MQTT:
           lack of explicit acknowledgements (in paho library) means ack's happen without the
           application being able to process the messages... potential for message loss of
           whatever is queued... this is not good for reliability.  
           It looks like just a choice in the paho library python version... other language bindings apparently have it.
           added to bug report: https://github.com/eclipse/paho.mqtt.python/pull/554
           made a pull request: https://github.com/eclipse/paho.mqtt.python/pull/554 
           Implemented support for the modification here. Seems to work fine (not thoroughly tested yet.)

      Why v5 only:

           No shared subscriptions prior to v5. Used extensively. had originally written library with v3
           support just restricting instance=1, but then I read about the re-transmission strategy:

           In MQTT < 5, in QoS==1, one is supposed to resend messages until you get an ack. some implementations
           do it every 20 seconds.  So in a case where you have, say 5 minutes of queueing, each message will be 
           sent 300/20 -> 15 times...  an unhelpful packet storm.


      There is additionally a vulnerability/inefficiency resulting from async message reception:
           when a new message arrive the loop thread started by the paho library will append
           it to the rx_msg data structure (protected by a mutex.)
           if the application crashes, the rx_msg have not been ack'd to the sender... so it is
           likely that you will just get them later... could use for a shelf for rx_msg,
           and sync it once in while... probably not worth it.
 
      Other:
           missing dry_run support.

      References:
         good discussion of shared subscriptions:
          https://www.hivemq.com/blog/mqtt5-essentials-part7-shared-subscriptions/#:~:text=Shared%20subscriptions%20are%20an%20MQTT,is%20sent%20to%20that%20topic.


    """
    def __init__(self, options, is_subscriber):
        """

          MQTT connection instances...
          in AMQP: topic separator is dot (.) in MQTT, the topic separator is slash (/)
          have exchange arguments use protocol specific separator or mapped one?

        """

        super().__init__(options, is_subscriber)

        self.connected=False
        # setting this is wrong and breaks things, was already set in super-class init, doing this here overwites
        #  interpretation of options done in superclass.
        #self.o = options
        self.o.update(default_options)
        self.o.update(options)

        if 'qos' in self.o:
            if type(self.o['qos']) is not int:
                self.o['qos'] = int(self.o['qos'])
        else:
            self.o['qos'] = 1


        me = "%s.%s" % (__class__.__module__, __class__.__name__)
        logger.setLevel('WARNING')

        if ('settings' in self.o) and (me in self.o['settings']):
            for s in self.o['settings'][me]:
                self.o[s] = self.o['settings'][me][s]

            if 'logLevel' in self.o['settings'][me]:
                logger.setLevel(self.o['logLevel'].upper())
        
        self.proto_version = paho.mqtt.client.MQTTv5

        if 'receiveMaximum' in self.o and type(self.o['receiveMaximum']) is not int:
            self.o['receiveMaximum'] = min( int( self.o['receiveMaximum'] ), 65535 )

        if 'max_inflight_messages' in self.o and type(self.o['max_inflight_messages']) is not int:
            self.o['max_inflight_messages'] = int( self.o['max_inflight_messages'] )

        if 'max_queued_messages' in self.o and type(self.o['max_queued_messages']) is not int:
            self.o['max_queued_messages'] = int( self.o['max_queued_messages'] )

        if is_subscriber:
            self.subscribe_mutex = threading.Lock()
            self.subscribe_mutex.acquire()
            self.subscribe_in_progress = 0
            self.subscribe_mutex.release()

            self.rx_msg_mutex = threading.Lock()
            self.rx_msg_mutex.acquire()
            self.rx_msg_iToApp=0
            self.rx_msg_iFromBroker=1
            self.rx_msg_iMax=4
            self.rx_msg={}
            self.rx_msg[0]=[]
            self.rx_msg[1]=[]
            self.rx_msg[2]=[]
            self.rx_msg[3]=[]
            self.rx_msg[4]=[]
            self.rx_msg_mutex.release()

      
        logger.warning("note: mqtt support is newish, not very well tested")

    def __sub_on_disconnect(client, userdata, rc, properties=None):
        userdata.metricsDisconnect()
        logger.debug(paho.mqtt.client.connack_string(rc))
        if hasattr(userdata, 'pending_publishes'):
            lost = len(userdata.pending_publishes)
            if lost > 0:
                logger.error(
                    'message loss! cannot confirm %d messages were published: mids=%s'
                    % (lost, userdata.pending_publishes))
            else:
                logger.info('clean. no published messages lost.')

    def __sub_on_connect(client, userdata, flags, rc, properties=None):
        logger.debug("client=%s rc=%s, flags=%s" %
                    (client, paho.mqtt.client.connack_string(rc), flags))

        if flags['session present'] != 1:
            logger.debug(
                'no existing session, no recovery of inflight messages from previous connection'
            )

        userdata.connect_in_progress = False
        if rc != paho.mqtt.client.MQTT_ERR_SUCCESS:
            logger.error("failed to establish connection")
            return

        # FIXME: enhancement could subscribe accepts multiple (subj, qos) tuples so, could do this in one RTT.
        userdata.subscribe_mutex.acquire()
        for binding_tuple in userdata.o['bindings']:

            if 'topic' in userdata.o:
                subj=userdata.o['topic']
            else:
                exchange, prefix, subtopic = binding_tuple
                logger.info("tuple: %s %s %s" % (exchange, prefix, subtopic))

                subj = '/'.join(['$share', userdata.o['queueName'], exchange] +
                                prefix + subtopic)

            (res, mid) = client.subscribe(subj, qos=userdata.o['qos'])
            userdata.subscribe_in_progress += 1
            logger.info( "asked to subscribe to: %s, mid=%d qos=%s result: %s" % (subj, mid, \
                userdata.o['qos'], paho.mqtt.client.error_string(res)) )
        userdata.subscribe_mutex.release()
        userdata.metricsConnect()


    def __sub_on_subscribe(client,
                           userdata,
                           mid,
                           granted_qos,
                           properties=None):
        userdata.subscribe_mutex.acquire()
        logger.info("client: {} subscribe completed mid={} granted_qos={}".format(
            client._client_id, mid, list(map(lambda x: x.getName(), granted_qos))))
        userdata.subscribe_in_progress -= 1
        userdata.subscribe_mutex.release()

    def __pub_on_disconnect(client, userdata, rc, properties=None):
        userdata.metricsDisconnect()
        logger.info(paho.mqtt.client.connack_string(rc))

    def __pub_on_connect(client, userdata, flags, rc, properties=None):
        userdata.connect_in_progress = False
        userdata.metricsConnect()
        logger.info(paho.mqtt.client.connack_string(rc))

    def __pub_on_publish(client, userdata, mid):

        if mid in userdata.pending_publishes:
            logger.info('publish complete. mid={}'.format(mid))
            # FIXME: worried... not clear if dequeue remove is thread safe.
            userdata.pending_publishes.remove(mid)
        else:
            userdata.unexpected_publishes.append(mid)
            logger.info( 'BUG: ack for message we do not know we published. mid={}'.  format(mid))

    def __sslClientSetup(self) -> int:
        """
          Initializse self.client SSL context, must be called after self.client is instantiated.
          return port number for connection.
      
        """
        if self.o['broker'].url.scheme[-1] == 's':
            port = 8883
            logger.info('tlsRigour: %s' % self.o['tlsRigour'])
            self.o['tlsRigour'] = self.o['tlsRigour'].lower()
            if self.o['tlsRigour'] == 'lax':
                self.tlsctx = ssl.create_default_context()
                self.tlsctx.check_hostname = False
                self.tlsctx.verify_mode = ssl.CERT_NONE

            elif self.o['tlsRigour'] == 'strict':
                self.tlsctx = ssl.SSLContext(ssl.PROTOCOL_TLS)
                self.tlsctx.options |= ssl.OP_NO_TLSv1
                self.tlsctx.options |= ssl.OP_NO_TLSv1_1
                self.tlsctx.check_hostname = True
                self.tlsctx.verify_mode = ssl.CERT_REQUIRED
                self.tlsctx.load_default_certs()
                # TODO Find a way to reintroduce certificate revocation (CRL) in the future
                #  self.tlsctx.verify_flags = ssl.VERIFY_CRL_CHECK_CHAIN
                #  https://github.com/MetPX/sarracenia/issues/330
            elif self.o['tlsRigour'] == 'normal':
                self.tlsctx = ssl.create_default_context()
            else:
                self.logger.warning(
                    "option tlsRigour must be one of:  lax, normal, strict")
            self.client.tls_set_context(self.tlsctx)
        else:
            port = 1883

        if self.o['broker'].url.port:
            port = self.o['broker'].url.port
        return port

    def __clientSetup(self, cid) -> paho.mqtt.client.Client:

        self.connect_in_progress = True

        if (self.o['broker'].url.scheme[-2:] == 'ws' ) or  \
           (self.o['broker'].url.scheme[-1] == 'w' ) : 
            client = paho.mqtt.client.Client( userdata=self, transport="websockets", \
                client_id=cid, protocol=paho.mqtt.client.MQTTv5 )
        else:
            client = paho.mqtt.client.Client( userdata=self, \
                client_id=cid, protocol=paho.mqtt.client.MQTTv5 )

        client.connected = False
        client.on_connect = MQTT.__sub_on_connect
        client.on_disconnect = MQTT.__sub_on_disconnect
        client.on_message = MQTT.__sub_on_message
        client.on_subscribe = MQTT.__sub_on_subscribe
        # defaults to 20... kind of a mix of "batch" and prefetch...
        if 'max_inflight_messages' in self.o:
            client.max_inflight_messages_set(self.o['max_inflight_messages'])

        client.username_pw_set(self.o['broker'].url.username,
                               unquote(self.o['broker'].url.password))
        return client

    def _mqtt_setup_signal_handler(self, signum, stack):
        logger.info("ok, asked to stop")
        self.please_stop=True

    def getSetup(self):
        """
           Establish a connection to consume messages with.  
        """
        ebo = 1
        original_sigint = signal.getsignal(signal.SIGINT)
        original_sigterm = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, self._mqtt_setup_signal_handler)
        signal.signal(signal.SIGTERM, self._mqtt_setup_signal_handler)

        something_broke = True
        self.connected=False
        self.manual_ack = False
        while True:

            if self.please_stop:
                break

            try:

                cs = self.o['clean_session']
                if ('queueName' in self.o) and ('no' in self.o):
                    cid = self.o['queueName'] + "_i%02d" % self.o['no']
                else:
                    #cid = ''.join(random.choice(string.ascii_lowercase) for i in range(10))
                    cid = None
                    cs = True
                    logger.info( f" cid=+{cid}+"  )

                props = Properties(PacketTypes.CONNECT)
                if 'expire' in self.o and self.o['expire']:
                    props.SessionExpiryInterval = int(self.o['expire'])
                if 'receiveMaximum' in self.o:
                    props.ReceiveMaximum = self.o['receiveMaximum']

                logger.info( f"is no around? {self.o['no']} " )
                if ('no' in self.o) and self.o['no'] > 0: # instances 'started'
                    self.client = self.__clientSetup(cid)
                    if hasattr(self, 'client') and hasattr(self.client, 'manual_ack'):  # FIXME breaking this...
                        self.client.manual_ack(True)
                        logger.debug(
                            "Switching on manual_acks for higher reliability via explicit acknowledgements."
                        )
                        self.manual_ack = True
                    else:
                        logger.warning(
                            "paho library automatically acknowledges receipt. may lose data every crash or restart."
                        )
    
                    self.client.connect( self.o['broker'].url.hostname, port=self.__sslClientSetup(), \
                           clean_start=False, properties=props )
                    self.client.enable_logger(logger)
                    self.client.loop_start()
                    ebo=1
                    while (self.connect_in_progress) or (self.subscribe_in_progress > 0):
                         logger.info( f"waiting for subscription to be set up. (ebo={ebo})")
                         time.sleep(0.1*ebo)
                         if self.please_stop:
                              break
                         if ebo < 512 :
                            ebo *= 2
                    self.connected=True
                    break
                else: # either 'declare' or 'foreground'
                    if 'instances' in self.o:    
                        session_mxi=self.o['instances']+1
                    else:
                        session_mxi=2

                    for i in range(1,session_mxi ):
                        icid = self.o['queueName'] + "_i%02d" %  i
                        logger.info('declare session for instances %s' %icid)
                        decl_client = self.__clientSetup(icid)
                        decl_client.on_connect = MQTT.__sub_on_connect
                        decl_client.connect( self.o['broker'].url.hostname, port=self.__sslClientSetup(), \
                           clean_start=False, properties=props )
                        while (self.connect_in_progress) or (self.subscribe_in_progress > 0):
                            decl_client.loop(1)
                        decl_client.disconnect()
                        decl_client.loop_stop()
                        logger.info('instance declaration for %s done' % icid)
                    break
                    
            except Exception as err:
                logger.error("failed to {} with {}".format(
                    self.o['broker'].url.hostname, err))
                logger.error('Exception details: ', exc_info=True)

            if ebo < 60: ebo *= 2
            time.sleep(ebo)

        signal.signal(signal.SIGINT, original_sigint)
        signal.signal(signal.SIGTERM, original_sigterm)
        if self.please_stop:
            signal.raise_signal(signal.SIGINT)



    def putSetup(self):
        """
           establish a connection to allow publishing. 
        """
        ebo = 1
        original_sigint = signal.getsignal(signal.SIGINT)
        original_sigterm = signal.getsignal(signal.SIGINT)
        signal.signal(signal.SIGINT, self._mqtt_setup_signal_handler)
        signal.signal(signal.SIGTERM, self._mqtt_setup_signal_handler)

        self.connected=False
        while True:
            
            if self.please_stop:
                break

            try:
                self.pending_publishes = collections.deque()
                self.unexpected_publishes = collections.deque()

                props = Properties(PacketTypes.CONNECT)
                if self.o['message_ttl'] > 0:
                    props.MessageExpiryInterval = int(self.o['message_ttl'])

                if (self.o['broker'].url.scheme[-2:] == 'ws' ) or  \
                   (self.o['broker'].url.scheme[-1] == 'w' ) : 
                    self.client = paho.mqtt.client.Client( userdata=self, transport="websockets", \
                        protocol=self.proto_version )
                else:
                    self.client = paho.mqtt.client.Client( userdata=self, \
                        protocol=self.proto_version )

                #self.client = paho.mqtt.client.Client(
                #    protocol=self.proto_version, userdata=self)

                self.client.enable_logger(logger)
                self.client.on_connect = MQTT.__pub_on_connect
                self.client.on_disconnect = MQTT.__pub_on_disconnect
                self.client.on_publish = MQTT.__pub_on_publish
                if 'max_queued_messages' in self.o:
                    self.client.max_queued_messages_set(self.o['max_queued_messages'])

                self.client.username_pw_set(self.o['broker'].url.username,
                                            unquote(self.o['broker'].url.password))
                self.connect_in_progress = True
                res = self.client.connect_async(self.o['broker'].url.hostname,
                                          port=self.__sslClientSetup(),
                                         properties=props)
                logger.info('connecting to %s, res=%s' % (self.o['broker'].url.hostname, res))

                self.client.loop_start()

                ebo=1
                while self.connect_in_progress:
                     logger.info( f"waiting for connection to {self.o['broker']} ebo={ebo}")
                     time.sleep(0.1*ebo)
                     if self.please_stop:
                          break
                     if ebo < 512:
                          ebo *= 2
                    

                if not self.connect_in_progress:
                    self.connected=True
                    break

            except Exception as err:
                logger.error("failed to {} with {}".format(
                    self.o['broker'].url.hostname, err))
                logger.error('Exception details: ', exc_info=True)

            if ebo < 60: ebo *= 2
            time.sleep(ebo)

        signal.signal(signal.SIGINT, original_sigint)
        signal.signal(signal.SIGTERM, original_sigterm)
        if self.please_stop:
            signal.raise_signal(signal.SIGINT)


    def __sub_on_message(client, userdata, msg):
        """
          callback to append messages received to new queue.

          FIXME: locking here is expensive... would like to group them ... 1st draft.
             could do a rotating set of batch size lists, and that way just change counters.
             and not lock individual messages  list[1], list[2], ... consumer consumes old lists...
             no locking needed, except to increment list index.... later.
        """

        if userdata.o['messageDebugDump']:
            logger.info("Message received: id:%d, topic:%s payload:%s" %
                        (msg.mid, msg.topic, msg.payload))

        m = userdata._msgDecode(msg)
        userdata.rx_msg[userdata.rx_msg_iFromBroker].append(m)

    def putCleanUp(self):
        self.client.disconnect()
        self.client.loop_stop()
        pass

    def getCleanUp(self):

        if not ('no' in self.o):
            props = Properties(PacketTypes.CONNECT)
            props.SessionExpiryInterval = 1
            for i in range(1,self.o['instances']+1):
                icid= self.o['queueName'] + "_i%02d" % i
                logger.info('cleanup session %s' % icid )
                myclient = self.__clientSetup( icid )
                myclient.connect( self.o['broker'].url.hostname, port=self.__sslClientSetup(), \
                   clean_start=False, properties=props )
                while self.connect_in_progress:
                    myclient.loop(0.1)
                    if self.please_stop:
                       break
                myclient.disconnect()
                logger.info('instance deletion for %02d done' % i)

        if hasattr(self, 'client'):
            self.client.disconnect()
            self.client.loop_stop()

    def _msgDecode(self, mqttMessage) -> sarracenia.Message:
        """
           decode MQTT message (protocol specific thingamabob) into sr3 one (python dictionary)
        """
        headers = { 'topic' : mqttMessage.topic }
        if self.o['messageDebugDump']:
            logger.info( f"raw message start topic={mqttMessage.topic}, qos={mqttMessage.qos}, mid={mqttMessage.mid}")
            if hasattr(mqttMessage.properties, 'ContentType'): 
                logger.info( f'Content-type: {mqttMessage.properties.ContentType}')
                headers['content-type'] = mqttMessage.properties.ContentType
            else:
                logger.warning('message is missing content-type header')

            if hasattr(mqttMessage, 'payload'): 
                logger.info('payload: type: %s (%d bytes) %s' %
                             (type(mqttMessage.payload), len(mqttMessage.payload), mqttMessage.payload))

            if hasattr(mqttMessage.properties, 'UserProperty'): 
                logger.info( f"User Property: {mqttMessage.properties.UserProperty} ")

        self.metrics['rxByteCount'] += len(mqttMessage.payload)
        try:
            if hasattr( mqttMessage.properties , 'UserProperty'):
                [ headers.update({k:v}) for k,v in mqttMessage.properties.UserProperty ]
            message = PostFormat.importAny( mqttMessage.payload.decode('utf-8'), headers, mqttMessage.properties.ContentType, self.o)

        except Exception as ex:
            logger.error("ignored malformed message: %s" % mqttMessage.payload)
            logger.error("decode error: %s" % ex)
            logger.error('Exception details: ', exc_info=True)
            self.metrics['rxBadCount'] += 1
            return None

        message['ack_id'] = mqttMessage.mid
        message['qos'] = mqttMessage.qos
        message['local_offset'] = 0
        message['_deleteOnPost'] |= set( ['exchange', 'local_offset', 'ack_id', 'qos' ])

        if message.validate():
            self.metrics['rxGoodCount'] += 1
            return message
        else:
           self.metrics['rxBadCount'] += 1
           self.ack(message)
           logger.error('message acknowledged and discarded: %s' % message)
           return None

    def _rotateInputBuffers(self) -> None:
        """
           to reduce locking granularity allocate one queue to accepting messages, and a second
           one to read from. and just swap between them from time to time (kind of double buffering.)

        """
        if len(self.rx_msg[self.rx_msg_iToApp]) == 0:
            self.rx_msg_mutex.acquire()
            self.rx_msg_iToApp = self.rx_msg_iFromBroker

            if self.rx_msg_iFromBroker >= self.rx_msg_iMax:
                self.rx_msg_iFromBroker=0
            else:
                self.rx_msg_iFromBroker+=1
            time.sleep(0.1)
            self.rx_msg_mutex.release()

    def newMessages(self) -> list:
        """
           return new messages.

           FIXME: hate the locking... too fine grained, especially in on_message... just a 1st shot.

        """

        #logger.debug( f"rx_msg queue before: indices: {self.rx_msg_iToApp} {self.rx_msg_iFromBroker} " )
        #logger.debug( f"rx_msg queue before: {len(self.rx_msg[self.rx_msg_iToApp])} indices: {self.rx_msg_iToApp} {self.rx_msg_iFromBroker} " )
        if not self.connected:
            self.getSetup()

        if len(self.rx_msg[self.rx_msg_iToApp]) > self.o['batch']:
            mqttml = self.rx_msg[self.rx_msg_iToApp][0:self.o['batch']]
            self.rx_msg[self.rx_msg_iToApp] = self.rx_msg[self.rx_msg_iToApp][self.o['batch']:]
        elif len(self.rx_msg[self.rx_msg_iToApp]) > 0:
            mqttml = self.rx_msg[self.rx_msg_iToApp]
            self.rx_msg[self.rx_msg_iToApp] = []
        else:
            mqttml = []

        #logger.debug( f"picked up {len(mqttml)} rx_msg queue after: {len(self.rx_msg[self.rx_msg_iToApp])} ")

        self._rotateInputBuffers()

        return mqttml

    def getNewMessage(self) -> sarracenia.Message:

        if not self.connected:
            self.getSetup()

        if len(self.rx_msg) > 0:
            m = self.rx_msg[self.rx_msg_iToApp][0]
            self.rx_msg[self.rx_msg_iToApp] = self.rx_msg[self.rx_msg_iToApp][1:]
        else:
            m = None
        self._rotateInputBuffers()

        if m:
            return m
        else:
            return None

    def ack(self, m: sarracenia.Message ) -> None:

        if self.manual_ack and ('ack_id' in m):
            logger.info('mid=%d' % m['ack_id'])
            self.client.ack( m['ack_id'], m['qos'] )
            del m['ack_id']
            m['_deleteOnPost'].remove('ack_id')

    def putNewMessage(self,
                      message: sarracenia.Message,
                      content_type: str ='application/json',
                      exchange: str = None ) -> bool:
        """
          publish a message.
        """
        if self.is_subscriber:  #build_consumer
            logger.error("publishing from a consumer")
            return False

        if not self.connected:
            self.putSetup()

        # The caller probably doesn't expect the message to get modified by this method, so use a copy of the message
        body = copy.deepcopy(message)

        postFormat = body['_format']

        if '_deleteOnPost' in body:
            # FIXME: need to delete because building entire JSON object at once.
            # makes this routine alter the message. Ideally, would use incremental
            # method to build json and _deleteOnPost would be a guide of what to skip.
            # library for that is jsonfile, but not present in repos so far.
            for k in body['_deleteOnPost']:
                if k in body:
                    del body[k]
            del body['_deleteOnPost']

        if not exchange:
            if (type(self.o['exchange']) is list):
                if (len(self.o['exchange']) > 1):
                    if 'post_exchangeSplit' in self.o:
                        # FIXME: assert ( len(self.o['exchange']) == self.o['post_exchangeSplit'] )
                        #        if that isn't true... then there is something wrong... should we check ?
                        idx = sum(
                            bytearray(body['identity']['value'],
                                      'ascii')) % len(self.o['exchange'])
                        exchange = self.o['exchange'][idx]
                    else:
                        logger.error(
                            'do not know which exchange to publish to: %s' %
                            self.o['exchange'])
                        return
                else:
                    exchange = self.o['exchange'][0]
            else:
                exchange = self.o['exchange']

        # https://docs.oasis-open.org/mqtt/mqtt/v5.0/os/mqtt-v5.0-os.html#_Toc3901111
        props = Properties(PacketTypes.PUBLISH)
        props.PayloadFormatIndicator = 1  # designates UTF-8

        props.ContentType = PostFormat.content_type( postFormat )

        try:
            raw_body, headers, content_type = PostFormat.exportAny( body, postFormat, self.o['topicPrefix'], self.o )
            topic = '/'.join(headers['topic']) 

            # url-quote wildcard characters in topics.
            topic = topic.replace('#', '%23')
            topic = topic.replace('+', '%2B')

            del headers['topic']

            if headers:
                props.UserProperty=list(map( lambda x :  (x,headers[x]) , headers ))

            if self.o['messageDebugDump']:
                logger.info( f"Message to publish: topic: {topic} body type:{type(raw_body)} body:{raw_body}" )
                if hasattr(props, 'UserProperty'): 
                    logger.info( f"UserProperty:{props.UserProperty}" )
                    

            info = self.client.publish(topic=topic, payload=raw_body, qos=self.o['qos'], properties=props)
               
            if info.rc == paho.mqtt.client.MQTT_ERR_SUCCESS:
                if info.mid in self.unexpected_publishes:
                    self.unexpected_publishes.remove(info.mid)
                    ack_pending=False
                else:
                    self.pending_publishes.append(info.mid)
                    ack_pending=True

                self.metrics['txByteCount'] += len(raw_body)
                self.metrics['txGoodCount'] += 1
                logger.info("published mid={} ack_pending={} {} to under: {} ".format(
                    info.mid, ack_pending, body, topic))
                return True  #success...
            logger.error( f"publish failed {paho.mqtt.client.error_string(info.rc)} ")

        except Exception as ex:
            logger.error('Exception details: ', exc_info=True)
            self.metrics['txBadCount'] += 1
            self.close()
        return False

    def close(self):
        logger.info('closing')
        if hasattr(self,'client') and self.client.is_connected():
            if self.is_subscriber and self.subscribe_in_progress:
                time.sleep(0.1)

            if hasattr(self, 'pending_publishes'):
                ebo=0.1
                while  (len(self.pending_publishes)+len(self.unexpected_publishes)) >0:
                    logger.info( f'waiting {ebo} seconds for last {len(self.pending_publishes)} messages to publish')
                    if len(self.unexpected_publishes) < 10:
                        logger.info( f'messages acknowledged before publish?: {self.unexpected_publishes}')
                    if len(self.pending_publishes) < 10:
                        logger.info( f'messages awaiting publish: {self.pending_publishes}')
                    time.sleep(ebo)
                    if ebo < 64:
                        ebo *= 2
                logger.info('no more pending messages')
            self.client.disconnect()
        self.connected=False



"""
Plugin tohttp.py:
    ToHttp is the converse of ToFile.
    After processing on a filter, a file URL needs to be turned back into a web url.
    Uses savedUrl created by ToFile, to convert file url back to original.

Usage:
    flowcb sarracenia.flowcb.accept.tohttp.ToHttp
"""
import logging
import re
import urllib

from sarracenia.flowcb import FlowCB
logger = logging.getLogger(__name__)


class ToHttp(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

        self._ldocroot = None
        
        #self.o.add_option('baseDir', 'str')
        if self.o.baseDir:
            self._ldocroot = self.o.baseDir

        self.o.add_option('toHttpRoot', 'str')
        if self.o.toHttpRoot:
            self._ldocroot = self.o.toHttpRoot

        #self.o.hurlre = re.compile('file:/' + self.o.ldocroot)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            logger.info("ToHttp message input: baseUrl=%s, relPath=%s" % (message['baseUrl'], message['relPath']))

            url = urllib.parse.urlparse(message['baseUrl'])

            new_baseUrl = 'http://' + url.netloc
            if self._ldocroot != None:
                new_baseUrl += self._ldocroot
            
            new_baseUrl += url.path

            message['baseUrl'] = new_baseUrl.replace('///', '//')

            logger.debug("ToHttp config; baseDir=%s, toHttpRoot=%s" % (self.o.baseDir, self.o.toHttpRoot))
            logger.info("ToHttp message output: baseUrl=%s, relPath=%s" % (message['baseUrl'], message['relPath']))



#!/usr/bin/env python3
"""
  assume directories are named
  YYYYMMDDHHTHH
 
  after the UTC date when the first product within it is received.
   
  this script looks at the list of directories and removes the ones older
  than requested.  Sample invocation:
 
   python3 old_hour_dirs.py "5 hours ago" /Project/web_root
 
"""

import dateparser, os, random, shutil, sys, time

how_many_hours = sys.argv[1]

try:
    os.chdir(sys.argv[2])
except:
    print("Second argument should directory I can visit")
    exit(2)

cutoff = dateparser.parse(how_many_hours)

print("last is: %s\n" % cutoff)

old_dirs = []

for d in os.listdir('.'):
    ddt = dateparser.parse(d, ['%Y%m%dT%H'])
    print("this one is: %s\n" % ddt)
    if ddt is not None and ddt < cutoff:
        old_dirs.append(d)
        continue
    print("skipping: %s" % d)

old_dirs.sort()
random.shuffle(old_dirs)
for d in old_dirs:

    while True:
        try:
            print("shutil.rmtree(%s)\n" % d)
            shutil.rmtree(d)

            if not os.path.isdir(d):
                print('succeeded? %s failed isdir' % d)
                break
            print("ugh. rmtree(%s) succeeded, but it is still there" % (d))
            time.sleep(30)
        except Exception as ex:
            print("ugh. rmtree(%s) failed: %s" % (d, ex))
            time.sleep(30)



#
# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, 2008-2021
#

import logging
import sarracenia
from sarracenia.flow import Flow
from sarracenia.featuredetection import features
import sys



logger = logging.getLogger(__name__)

default_options = {
    'acceptUnmatched': True,
    'blocksize': 1,
    'bufsize': 1024 * 1024,
    'chmod': 0o400,
    'pollUrl': None,
    'follow_symlinks': False,
    'force_polling': False,
    'inflight': None,
    'identity_method': 'cod,sha512',
    'part_ext': 'Part',
    'partflg': '1',
    'post_baseDir': None,
    'permCopy': True,
    'timeCopy': True,
    'randomize': False,
    'post_on_start': False,
    'sleep': -1,
    'nodupe_ttl': 7 * 60 * 60,
    'nodupe_fileAgeMax': 30 * 24 * 60 * 60,
}

#  'sumflg': 'cod,md5',


class Poll(Flow):
    """
       repeatedly query a remote (non-sarracenia) server to list the files there.
       post messages (to post_broker) for every new file discovered there.

       the sarracenia.flowcb.poll class is used to implement the remote querying,
       and is highly customizable to that effect.

       if the vip option is set, 
       * subscribe to the same settings that are being posted to.
       * consume all the messages posted, keeping new file duplicate cache updated.
          
    """
    def __init__(self, options):

        super().__init__(options)

        if hasattr(self.o,'post_exchange') and hasattr(self.o,'exchange'):
            px = self.o.post_exchange if type(self.o.post_exchange) != list else self.o.post_exchange[0]
            if px != self.o.exchange:
                logger.warning( f"post_exchange: {px} is different from exchange: {self.o.exchange}. The settings need for multiple instances to share a poll." )
            else:
                logger.info( f"Good! post_exchange: {px} and exchange: {self.o.exchange} match so multiple instances to share a poll." )

        if not 'poll' in ','.join(self.plugins['load']):
            logger.info( f"adding poll plugin, because missing from: {self.plugins['load']}" ) 
            self.plugins['load'].append('sarracenia.flowcb.poll.Poll')

        if options.vip:
            self.plugins['load'].insert(
                0, 'sarracenia.flowcb.gather.message.Message')

        self.plugins['load'].insert(0,
                                    'sarracenia.flowcb.post.message.Message')

        if not features['ftppoll']['present']:
            if hasattr( self.o, 'pollUrl' ) and ( self.o.pollUrl.startswith('ftp') ):
                logger.critical( f"attempting to configure an FTP poll pollUrl={self.o.pollUrl}, but missing python modules: {' '.join(features['ftppoll']['modules_needed'])}" )
                sys.exit(1)

    def do(self):
        """
            stub to do the work: does nothing, marking everything done.
            to be replaced in child classes that do transforms or transfers.
        """

        # mark all remaining messages as rejected.
        if self.worklist.poll_catching_up:
            # in catchup mode, just reading previously posted messages.
            self.worklist.rejected = self.worklist.incoming
        else:
            self.worklist.ok = self.worklist.incoming

        logger.debug('processing %d messages worked! (stop requested: %s)' %
                     (len(self.worklist.incoming), self._stop_requested))
        self.worklist.incoming = []


    def gather(self):

        super().gather()

        if len(self.worklist.incoming) > 0:
            logger.info('ingesting %d postings into duplicate suppression cache' % len(self.worklist.incoming) )
            self.worklist.poll_catching_up = True
            return 
        else:
            self.worklist.poll_catching_up = False

        if self.have_vip:
            for plugin in self.plugins['poll']:
                new_incoming = plugin()
                if len(new_incoming) > 0:
                    self.worklist.incoming.extend(new_incoming)




"""
Place holder customized flow callback classes whose focus is the download() entry point.

download(self,msg):

    Task: looking at msg['new_dir'], msg['new_path'], msg['new_inflight_path'] 
          and the self.o options perform a download of a single file.
          return True on a successful transfer, False otherwise.

    if self.o.dry_run is set, simulate the output of a download without
    performing it.

    downlaod is called by the sarra and subscribe components, and can be used 
    to override the built-in methods for downloading a file. 

    It does the download based on the fields in the message provided:

    retreival path: msg['baseUrl'] + '/' + msg['relPath'] 

    taking care of the inflight setting, the inflight/temporary file: 
    is defined for use by the download routing: msg['new_inflight_path'] 

    Final local location to store is defined by:

        msg['new_path'] == msg['new_dir'] + '/' + msg['new_file']

 
    This replaces built-in download functionality, providing an override.
    for individual file transfers. ideally you set checksums as you download.
            
    looking at self.o.identity_method to establish download checksum algorithm.
    might have to allow for cod... say it is checksum_method::
            
         checksum = sarracenia.identity.Identity.factory(self.o.checksum_method)
         while downloading:
             checksum.update(chunk)

    where chunk is the bytes read. It saves a file read to calculate the checksum
    during the download.
 
    if the checksum does not match what is in the received message, then 
    it is imperative, to avoid looping, to apply the actual checksum of the
    data to the message:

         msg['identity'] =  { 'method': checksum_method, 'value': checksum.get_sumstr() }
   
    return Boolean success indicator.  if False, download  will be attempted again and/or
    appended to retry queue.


"""

pass



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, 2008-2021
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
#

import logging, paramiko, os, subprocess, sys, time
from paramiko import *
from stat import *

from sarracenia.transfer import Transfer
from sarracenia.transfer import alarm_cancel, alarm_set, alarm_raise
from urllib.parse import unquote

import logging

logger = logging.getLogger(__name__)


class Sftp(Transfer):
    """
    SecSH File Transfer Protocol (SFTP)  ( https://filezilla-project.org/specs/draft-ietf-secsh-filexfer-02.txt )
    Sarracenia transfer protocol subclass supports/uses additional custom options:

    * accelScpCommand (default: '/usr/bin/scp %s %d' )

    The module uses the paramiko library for python SecSH support ( https://www.paramiko.org/ )
    """
    def __init__(self, proto, options):

        super().__init__(proto, options)

        logger.debug("sr_sftp __init__")

        self.o.add_option("accelScpCommand", "str", "/usr/bin/scp %s %d")
        # sftp command times out after 20 secs
        # this setting is different from the computed timeout (protocol)

        self.connected = False
        self.sftp = None
        self.ssh = None
        self.seek = True

        self.batch = 0
        self.connected = False
        self.ssh_config = None

        try:
            self.ssh_config = paramiko.SSHConfig()
            ssh_config = os.path.expanduser('~/.ssh/config')
            if os.path.isfile(ssh_config):
                fp = open(ssh_config, 'r')
                self.ssh_config.parse(fp)
                fp.close()
        except:
            logger.error("sr_sftp/__init__: unable to load ssh config %s" %
                         ssh_config)
            logger.debug('Exception details: ', exc_info=True)

    def registered_as():
        return ['sftp', 'scp', 'ssh', 'fish']

    # cd
    def cd(self, path):
        logger.debug("sr_sftp cd %s" % path)
        alarm_set(self.o.timeout)
        self.sftp.chdir(self.originalDir)
        self.sftp.chdir(path)
        self.pwd = path
        alarm_cancel()

    # cd forced
    def cd_forced(self, perm, path):
        logger.debug("sr_sftp cd_forced %d %s" % (perm, path))

        # try to go directly to path

        alarm_set(self.o.timeout)
        self.sftp.chdir(self.originalDir)
        try:
            self.sftp.chdir(path)
            alarm_cancel()
            return
        except:
            pass
        alarm_cancel()

        # need to create subdir

        subdirs = path.split("/")
        if path[0:1] == "/": subdirs[0] = "/" + subdirs[0]

        for d in subdirs:
            if d == '': continue
            # try to go directly to subdir
            try:
                alarm_set(self.o.timeout)
                self.sftp.chdir(d)
                alarm_cancel()
                continue
            except:
                pass

            # create and go to subdir
            alarm_set(self.o.timeout)
            self.sftp.mkdir(d, self.o.permDirDefault)
            self.sftp.chdir(d)
            alarm_cancel()

    def check_is_connected(self):
        logger.debug("sr_sftp check_is_connected")

        if self.sftp == None: return False
        if not self.connected: return False

        if self.sendTo != self.o.sendTo:
            self.close()
            return False

        self.batch = self.batch + 1
        if self.batch > self.o.batch:
            self.close()
            return False

        # really connected, getcwd would not work, send_ignore would not work... so chdir used
        try:
            alarm_set(self.o.timeout)
            self.sftp.chdir(self.originalDir)
            alarm_cancel()
        except:
            self.close()
            return False

        return True

    # chmod
    def chmod(self, perm, path):
        logger.debug("sr_sftp chmod %s %s" % ("{0:o}".format(perm), path))
        alarm_set(self.o.timeout)
        self.sftp.chmod(path, perm)
        alarm_cancel()

    # close
    def close(self):
        logger.debug("sr_sftp close")

        old_sftp = self.sftp
        old_ssh = self.ssh

        self.init()

        alarm_set(self.o.timeout)
        try:
            old_sftp.close()
        except:
            pass
        try:
            old_ssh.close()
        except:
            pass
        alarm_cancel()

    # connect...
    def connect(self):

        logger.debug("sr_sftp connect %s" % self.o.sendTo)

        if self.connected: self.close()

        self.connected = False
        self.sendTo = self.o.sendTo

        if not self.credentials(): return False

        alarm_set(self.o.timeout)
        try:

            sublogger = logging.getLogger('paramiko')
            sublogger.setLevel(logging.CRITICAL)
            self.ssh = paramiko.SSHClient()
            # FIXME this should be an option... for security reasons... not forced
            self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            if self.password:
                self.ssh.connect(self.host,self.port,self.user,unquote(self.password), \
                                 pkey=None,key_filename=self.ssh_keyfile,\
                                 timeout=self.o.timeout,allow_agent=False,look_for_keys=False)
            else:
                self.ssh.connect(self.host,self.port,self.user,self.password, \
                                 pkey=None,key_filename=self.ssh_keyfile,\
                                 timeout=self.o.timeout)
            #if ssh_keyfile != None :
            #  key=DSSKey.from_private_key_file(ssh_keyfile,password=None)

            sftp = self.ssh.open_sftp()
            if self.o.timeout != None:
                logger.debug("sr_sftp connect setting timeout %f" %
                             self.o.timeout)
                channel = sftp.get_channel()
                channel.settimeout(self.o.timeout)

            sftp.chdir('.')
            self.originalDir = sftp.getcwd()
            self.pwd = self.originalDir

            self.connected = True
            self.sftp = sftp

            #alarm_cancel()
            return True

        except:
            logger.error("sr_sftp/connect: unable to connect to %s (user:%s)" %
                         (self.host, self.user))
            logger.debug('Exception details: ', exc_info=True)

        alarm_cancel()
        return False

    # credentials...
    def credentials(self):
        #logger.debug("sr_sftp credentials %s" % self.sendTo)

        try:
            ok, details = self.o.credentials.get(self.sendTo)
            if details: url = details.url

            self.host = url.hostname
            self.port = url.port
            self.user = url.username
            self.password = url.password
            self.ssh_keyfile = details.ssh_keyfile

            if url.username == '': self.user = None
            if url.password == '': self.password = None
            if url.port == '': self.port = None
            if self.ssh_keyfile: self.password = None

            if self.port == None: self.port = 22

            #logger.debug("h u:p s = %s:%d %s:%s %s" %
            #             (self.host, self.port, self.user, self.password,
            #              self.ssh_keyfile))

            if self.ssh_config == None: return True

            if self.user        == None or \
             ( self.ssh_keyfile == None and self.password == None):
                #logger.debug("check in ssh_config")
                for key, value in self.ssh_config.lookup(self.host).items():
                    if key == "hostname":
                        self.host = value
                    elif key == "user":
                        self.user = value
                    elif key == "port":
                        self.port = int(value)
                    elif key == "identityfile":
                        self.ssh_keyfile = os.path.expanduser(value[0])

            #logger.debug("h u:p s = %s:%d %s:%s %s" %
            #             (self.host, self.port, self.user, self.password,
            #              self.ssh_keyfile))
            return True

        except:
            logger.error(
                "sr_sftp/credentials: unable to get credentials for %s" %
                self.sendTo)
            logger.debug('Exception details: ', exc_info=True)

        return False

    # delete
    # MG sneak rmdir here in case 'R' message implies a directory (remote mirroring)
    def delete(self, path):
        logger.debug("sr_sftp rm %s" % path)

        alarm_set(self.o.timeout)
        # check if the file is there... if not we are done,no error
        try:
            s = self.sftp.lstat(path)
        except:
            alarm_cancel()
            return

        # proceed with file/link removal
        if not S_ISDIR(s.st_mode):
            logger.debug("sr_sftp remove %s" % path)
            self.sftp.remove(path)

        # proceed with directory removal
        else:
            logger.debug("sr_sftp rmdir %s" % path)
            self.sftp.rmdir(path)

        alarm_cancel()

    def readlink(self, link):
        logger.debug("%s" % (link))
        alarm_set(self.o.timeout)
        value = self.sftp.readlink(link)
        alarm_cancel()
        return value

    # symlink
    def symlink(self, link, path):
        logger.debug("(in %s), create this file %s as a link to: %s" % (self.getcwd(), path, link) )
        alarm_set(self.o.timeout)
        self.sftp.symlink(link, path)
        alarm_cancel()

    # get

    def get(self,
            msg,
            remote_file,
            local_file,
            remote_offset=0,
            local_offset=0,
            length=0):
        logger.debug(
            "sr_sftp get %s %s %d %d %d" %
            (remote_file, local_file, remote_offset, local_offset, length))

        alarm_set(2 * self.o.timeout)
        rfp = self.sftp.file(remote_file, 'rb', self.o.bufsize)
        if remote_offset != 0: rfp.seek(remote_offset, 0)
        rfp.settimeout(1.0 * self.o.timeout)
        alarm_cancel()

        # read from rfp and write to local_file

        rw_length = self.read_writelocal(remote_file, rfp, local_file,
                                         local_offset, length)

        
        # close

        alarm_set(self.o.timeout)
        rfp.close()
        alarm_cancel()

        return rw_length

    def getAccelerated(self, msg, remote_file, local_file, length=0):

        base_url = msg['baseUrl'].replace('sftp://', '')
        if base_url[-1] == '/':
            base_url = base_url[0:-1]
        arg1 = base_url + ':' + self.pwd + os.sep + remote_file
        arg1 = arg1.replace(' ', '\ ')
        arg2 = local_file

        cmd = self.o.accelScpCommand.replace('%s', arg1)
        cmd = cmd.replace('%d', arg2).split()
        logger.info("accel_sftp:  %s" % ' '.join(cmd))
        p = subprocess.Popen(cmd)
        p.wait()
        if p.returncode != 0:
            return -1
        sz = os.stat(arg2).st_size
        return sz

    # getcwd
    def getcwd(self):
        alarm_set(self.o.timeout)
        cwd = self.sftp.getcwd() if self.sftp else None
        alarm_cancel()
        return cwd

    # ls
    def ls(self):
        logger.debug("sr_sftp ls")
        self.entries = {}
        # timeout is at least 30 secs, say we wait for max 5 mins
        alarm_set(self.o.timeout)
        dir_attr = self.sftp.listdir_attr()
        alarm_cancel()
        for index in range(len(dir_attr)):
            attr = dir_attr[index]
            line = attr.__str__()
            self.line_callback(line, attr)
        #logger.debug("sr_sftp ls = %s" % self.entries )
        return self.entries

    # line_callback: ls[filename] = 'stripped_file_description'
    def line_callback(self, iline, attr):
        #logger.debug("sr_sftp line_callback %s" % iline)

        oline = iline
        oline = oline.strip('\n')
        oline = oline.strip()
        oline = oline.replace('\t', ' ')
        opart1 = oline.split(' ')
        opart2 = []

        for p in opart1:
            if p == '': continue
            opart2.append(p)
        # else case is in the event of unlikely race condition
        fil = ' '.join(opart2[8:])
        line = ' '.join(opart2)

        self.entries[fil] = attr

    # mkdir
    def mkdir(self, remote_dir):
        logger.debug("mkdir %s" % remote_dir)
        alarm_set(self.o.timeout)
        
        try:
            s = self.sftp.lstat(path)
            if S_ISDIR(s.st_mode):
                return
            logger.error( f"cannot mkdir {path}, file exists" )
            alarm_cancel()
            return
        except FileNotFoundError:
            pass
        except:
            alarm_cancel()
            return

        self.sftp.mkdir(remote_dir, self.o.permDirDefault)
        alarm_cancel()

    # put
    def put(self,
            msg,
            local_file,
            remote_file,
            local_offset=0,
            remote_offset=0,
            length=0):
        logger.debug(
            "sr_sftp put %s %s %d %d %d" %
            (local_file, remote_file, local_offset, remote_offset, length))

        # simple file

        alarm_set(2 * self.o.timeout)

        if length == 0:
            rfp = self.sftp.file(remote_file, 'wb', self.o.bufsize)
            rfp.settimeout(1.0 * self.o.timeout)

        # parts
        else:
            try:
                self.sftp.stat(remote_file)
            except:
                rfp = self.sftp.file(remote_file, 'wb', self.o.bufsize)
                rfp.close()

            rfp = self.sftp.file(remote_file, 'r+b', self.o.bufsize)
            rfp.settimeout(1.0 * self.o.timeout)
            if remote_offset != 0: rfp.seek(remote_offset, 0)

        alarm_cancel()

        # read from local_file and write to rfp

        rw_length = self.readlocal_write(local_file, local_offset, length, rfp)

        # no sparse file... truncate where we are at

        alarm_set(self.o.timeout)
        self.fpos = remote_offset + rw_length
        if length != 0: rfp.truncate(self.fpos)
        rfp.close()
        alarm_cancel()

        return rw_length

    def putAccelerated(self, msg, local_file, remote_file, length=0):

        dest_baseUrl = self.o.sendTo.replace('sftp://', '')
        if dest_baseUrl[-1] == '/':
            dest_baseUrl = dest_baseUrl[0:-1]
        arg2 = dest_baseUrl + ':' + msg['new_dir'] + os.sep + remote_file
        arg2 = arg2.replace(' ', '\ ')
        arg1 = local_file

        cmd = self.o.accelScpCommand.replace('%s', arg1)
        cmd = cmd.replace('%d', arg2).split()

        logger.info("accel_sftp:  %s" % ' '.join(cmd))
        p = subprocess.Popen(cmd)
        p.wait()
        if p.returncode != 0:
            return -1
        # FIXME: faking success... not sure how to check really.
        sz = int(msg['size'])
        return sz

    # rename
    def rename(self, remote_old, remote_new):
        logger.debug("sr_sftp rename %s %s" % (remote_old, remote_new))
        try:
            self.delete(remote_new)
        except:
            pass
        alarm_set(self.o.timeout)
        self.sftp.rename(remote_old, remote_new)
        alarm_cancel()

    # rmdir
    def rmdir(self, path):
        logger.debug("sr_sftp rmdir %s " % path)
        alarm_set(self.o.timeout)
        self.sftp.rmdir(path)
        alarm_cancel()

    # utime
    def utime(self, path, tup):
        logger.debug("sr_sftp utime %s %s " % (path, tup))
        alarm_set(self.o.timeout)
        self.sftp.utime(path, tup)
        alarm_cancel()



from sarracenia.flow import Flow
import logging
import urllib.parse

logger = logging.getLogger(__name__)

default_options = {'acceptUnmatched': True, 'download': True}


class Sender(Flow):
    """
       * subscribe to a stream of messages about local files.
       * send the files to a remote server.
       * modify the messages to refer to the remote file copies.
       * post the messages for subscribers of the remote server.
    """
    def __init__(self, options):

        super().__init__(options)
        self.plugins['load'].insert(
            0, 'sarracenia.flowcb.gather.message.Message')

        if hasattr(self.o, 'post_exchange'):
            self.plugins['load'].insert(
                0, 'sarracenia.flowcb.post.message.Message')

        self.scheme = urllib.parse.urlparse(self.o.sendTo).scheme

    def do(self):

        self.do_send()

        logger.debug('processing %d messages worked!' % len(self.worklist.ok))



from base64 import b64decode, b64encode
from codecs import decode, encode

from sarracenia.flowcb import v2wrapper

import logging
import sarracenia
from sarracenia.postformat import PostFormat

logger = logging.getLogger(__name__)


class V02(PostFormat):
    """
       A class for controlling the format of messages are sent.
       internally All messages are represented as v03. 
       post format implement translations to other protocols for interop.

   """

    @staticmethod
    def content_type():
        return 'text/plain'

    @staticmethod
    def mine(payload, headers, content_type, options) -> bool:
        """
          return true if the message is in this post format.
       """
        if content_type == V02.content_type() :
            return True
        return False

    @staticmethod
    def importMine(body, headers, options) -> sarracenia.Message:
        """
          given a message in a wire format, with the given properties (or headers) in a dictionary,
          return the message as a normalized v03 message.

          the topic and topicPrefix should be lists, not protocol specific strings.
       """
        msg = sarracenia.Message()
        msg["_format"] = __name__.split('.')[-1].lower()
        msg.copyDict(headers)
    
        try:
            pubTime, baseUrl, relPath = body.split(' ')[0:3]
        except Exception as ex:
            logger.error( f"body should have three space separated fields: {body}, error: {ex}" )
            return None

        msg['pubTime'] = sarracenia.timev2tov3str(pubTime)
        msg['baseUrl'] = baseUrl.replace('%20', ' ').replace('%23', '#')
        msg['relPath'] = relPath
        msg['subtopic'] = relPath.split('/')
        msg['to_clusters'] = 'ALL'
        msg['_deleteOnPost'] |= set(['subtopic'])

        for t in ['atime', 'mtime']:
            if t in msg:
                try: 
                    msg[t] = sarracenia.timev2tov3str(msg[t])
                except Exception as ex:
                    logger.warning( f"invalid time field: {t} value: {msg['t']}, error: {ex}" )
        try:
            if 'sum' in msg:
                sum_algo_map = {
                    "0": "random",
                    "a": "arbitrary",
                    "d": "md5",
                    "L": "link",
                    "m": "mkdir",
                    "n": "md5name",
                    "r": "rmdir",
                    "R": "remove",
                    "s": "sha512",
                    "z": "cod"
                }
                sm = sum_algo_map[msg["sum"][0]]
                if sm in ['random', 'arbitrary']:
                    sv = msg["sum"][2:]
                elif sm in ['cod']:
                    sv = sum_algo_map[msg["sum"][2:]]
                else:
                    sv = encode(decode(msg["sum"][2:], 'hex'),
                                'base64').decode('utf-8').strip()
                if 'oldname' in msg:
                    msg['fileOp'] = { 'rename': msg['oldname'] }
                    del msg['oldname']
                    if sm in ['mkdir']:
                        msg['fileOp']['mkdir'] = ''
                    elif sm in ['link']:
                        msg['fileOp']['link'] = msg['link']
                    else:
                        msg["identity"] = {"method": sm, "value": sv}
                elif sm == 'remove':
                    msg['fileOp'] = { 'remove': '' }
                elif sm == 'mkdir':
                    msg['fileOp'] = { 'directory': '' }
                elif sm == 'rmdir':
                    msg['fileOp'] = { 'remove':'', 'directory': '' }
                elif 'link' in msg:
                    msg['fileOp'] = { 'link': msg['link'] }
                    del msg['link']
                elif sm == 'md5name':
                    pass
                else:
                    msg["identity"] = {"method": sm, "value": sv}
    
                del msg['sum']
        except Exception as ex:
            logger.warning( f"sum field corrupt: {msg['sum']} ignored: {ex}" )

        try:
            if 'parts' in msg:
                (style, chunksz, block_count, remainder,
                 current_block) = msg['parts'].split(',')
                if style in ['i', 'p']:
                    msg['blocks'] = {}
                    msg['blocks']['method'] = {
                        'i': 'inplace',
                        'p': 'partitioned'
                    }[style]
                    msg['blocks']['size'] = int(chunksz)
                    msg['blocks']['count'] = int(block_count)
                    msg['blocks']['remainder'] = int(remainder)
                    msg['blocks']['number'] = int(current_block)
                else:
                    msg['size'] = int(chunksz)
                del msg['parts']
        except Exception as ex:
            logger.warning( f"parts field corrupt: {msg['parts']}, ignored: {ex}" )
    
        return msg

    @staticmethod
    def exportMine(body, options) -> (str, dict, str):
        """
           given a v03 (internal) message, produce an encoded version.
       """
        v2m = v2wrapper.Message(body)
                                
        # v2wrapp
        for h in [
                    'pubTime', 'baseUrl', 'fileOp', 'relPath', 'size', 
                    'blocks', 'content', 'identity'
        ]:
            if h in v2m.headers:
                    del v2m.headers[h]

        v2m.headers['topic'] = PostFormat.topicDerive( body, options )

        return v2m.notice, v2m.headers, V02.content_type()



"""
Plugin wmotypesuffix.py:
    Given the WMO-386 TT designator of a WMO file, file type suffix to the file name.
    Web browsers and modern operating systems may do *the right thing* if files have a recognizable suffix.

    http://www.wmo.int/pages/prog/www/ois/Operational_Information/Publications/WMO_386/AHLsymbols/TableA.html

    Status: proof of concept demonstrator... missing many TT's. please add!
    Tested with UNIDATA feeds, discrepancies:
    TableA says L is Aviation XML, but UNIDATA Feed, it is all GRIB.
    XW - should be CAP, but is GRIB.
    IX used by Americans for HDF, unsure if that is kosher/halal/blessed, but it is in the UNIDATA feed.

    IU/IS/IB are BUFR
    other type designators welcome... for example, GRIB isn't identified yet.
    default to .txt.

Usage:
    flowcb sarracenia.flowcb.accept.wmotypesuffix.WmoTypeSuffix
  
"""
import logging
from sarracenia.flowcb import FlowCB
logger = logging.getLogger(__name__)


class WmoTypeSuffix(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

    def __find_type(self, TT):
        if TT[0] in ['G']: return '.grid'
        if TT in ['IX']: return '.hdf'
        if TT[0] in ['I']: return '.bufr'
        if TT[0] in ['K']: return '.crex'
        if TT in ['LT']: return '.iwxxm'
        if TT[0] in ['L']: return '.grib'
        if TT in ['XW']: return '.txt'
        if TT[0] in ['X']: return '.cap'
        if TT[0] in ['D', 'H', 'O', 'Y']: return '.grib'
        if TT[0] in ['E', 'P', 'Q', 'R']: return '.bin'
        return '.txt'

    def after_accept(self, worklist):
        for message in worklist.incoming:
            type_suffix = self.__find_type(message['new_file'][0:2])
            ## FIXME confused as to how this could ever be true since find_type never returns "UNKNOWN"
            #if type_suffix == 'UNKNOWN':
            #    continue

            # file name already has suffix
            if message['new_file'][-len(type_suffix):] == type_suffix:
                continue

            message['new_file'] = message['new_file'] + type_suffix
            if 'rename' in message:
                message['rename'] = message['rename'] + type_suffix
            # TODO else -> worklist.rejected.append(message) ?? should this be happening at any point?



import logging
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class Name(FlowCB):
    """
      Override the the comparison so that files with the same name,
      regardless of what directory they are in, are considered the same.
      This is useful when receiving data from two different sources (two different trees)
      and winnowing between them.
    """
    def after_accept(self, worklist):
        for m in worklist.incoming:
            if not 'nodupe_override' in m:
                m['_deleteOnPost'] |= set(['nodupe_override'])
                m['nodupe_override'] = {}

            m['nodupe_override']['path'] = m['relPath'].split('/')[-1]
            m['nodupe_override']['key'] = m['relPath'].split('/')[-1]



"""
Plugin testretry.py:
    This changes the message randomly so that it will cause
    a download or send an error.
    When a message is being retried, it is randomly fixed

Usage:
    flowcb sarracenia.flowcb.accept.testretry.TestRetry
"""

import logging
import random
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class TestRetry(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)
        self.sendTo = None
        self.msg_baseUrl_good = None
        self.details_bad = None
        self.msg_baseUrl_bad = 'sftp://ruser:rpass@retryhost'

    def after_accept(self, worklist):
        new_incoming = []
        for message in worklist.incoming:
            logger.debug("testretry after_accept")

            if self.sendTo == None:
                self.sendTo = self.o.sendTo
            if self.msg_baseUrl_good == None:
                self.msg_baseUrl_good = message['baseUrl']

            # retry message : recover it
            # update: this is set somewhere as true in /diskqueue.py, should think about initializing first so we
            # dont have to test for existence
            # 2022-06-10: isRetry was removed. Maybe can check if the message has msg_baseUrl_bad?
            #             see issues #466 and #527.
            if 'isRetry' in message and message['isRetry']:
                self.o.sendTo = self.sendTo
                ok, self.o.details = self.o.credentials.get(self.sendTo)

                ## # FIXME dont see 'set_notice' as an entry in the message dictionary, could cause an error
                ## #message['set_notice'](self.msg_baseUrl_good, message['relPath'], message['pubTime'])
                # Fixed missing message.set_notice method; now just set baseUrl
                message['baseUrl'] = self.msg_baseUrl_good

            # original message :  50% chance of breakage
            elif random.randint(0, 2):

                # if sarra or subscribe break download
                if self.o.component != 'sender':
                    logger.debug("making it bad 1")
                    ok, self.o.details = self.o.credentials.get(self.sendTo)

                    ## # FIXME dont see 'set_notice' as an entry in the message dictionary, could cause an error
                    ## #message['set_notice'](self.msg_baseUrl_bad, message['relpath'], message['pubTime'])
                    # Fixed missing message.set_notice method; now just set baseUrl
                    message['baseUrl'] = self.msg_baseUrl_bad

                # if sender break destination
                else:
                    logger.debug("making it bad 2")
                    self.o.sleep_connect_try_interval_max = 1.0
                    self.o.sendTo = self.msg_baseUrl_bad
                    self.o.credentials._parse(self.msg_baseUrl_bad)
                    ok, self.o.details = self.o.credentials.get(self.msg_baseUrl_bad)

            logger.debug("return from testretry after_accept")
            # TODO not sure where to add to new_incoming. as of now not appending to new_incoming or worklist.rejected
        worklist.incoming = new_incoming



"""
a module that is focused on transforming data should be called a filter.
Filter plugins intended for after_accept(self, worklist) entry_point.

At that point:
Messages have been gathered, then passed through the accept and reject pattern matches.
One of the first callbacks is the nodupe, so that duplicate suppression may cause additional
rejections.

"""
pass



from hashlib import md5

from sarracenia.identity import Identity


class Md5(Identity):
    """
         use the (obsolete) Message Digest 5 (MD5) algorithm, applied on the content
         of a file, to generate an identity signature.
      """

    @staticmethod
    def registered_as():
        """
            v2name.
          """
        return 'd'

    def set_path(self, path):
        self.filehash = md5()

    def update(self, chunk):
        if type(chunk) == bytes: self.filehash.update(chunk)
        else: self.filehash.update(bytes(chunk, 'utf-8'))



from base64 import b64decode, b64encode
from codecs import decode, encode

import copy
import logging
from hashlib import md5
from hashlib import sha512
import os
import sarracenia.config
import time
import types
import urllib

import sarracenia
from sarracenia.flowcb import FlowCB

from sarracenia import nowflt, timestr2flt, timev2tov3str

logger = logging.getLogger(__name__)

sum_algo_v3tov2 = {
                "arbitrary": "a",
                "md5": "d",
                "sha512": "s",
                "md5name": "n",
                "random": "0",
                "link": "L",
                "remove": "R",
                "cod": "z"
}

sum_algo_v2tov3 = { v: k for k,v in sum_algo_v3tov2.items() }

def sumstrFromMessage( msg ) -> str:
    """
   accepts a v3 message as argument msg. returns the corresponding sum string for a v2 'sum' header.
    """

    if 'identity' in msg:
        if msg['identity']['method'] in sum_algo_v3tov2:
           sa = sum_algo_v3tov2[msg["identity"]["method"]]
        else: # FIXME ... 1st md5name case... default when unknown...
           logger.error('identity method unknown to v2: %s, replacing with md5name' % msg['identity']['method'] )
           sa = 'n'
           sv = md5(bytes(os.path.basename(msg['relPath']),'utf-8')).hexdigest()

        # transform sum value
        if sa in ['0', 'a']:
            sv = msg["identity"]["value"]
        elif sa in ['z']:
            sv = sum_algo_v3tov2[msg["identity"]["value"]]
        else:
            sv = encode(
                decode(msg["identity"]["value"].encode('utf-8'), "base64"),
                'hex').decode('utf-8')
        sumstr = sa + ',' + sv
    else:
        # FIXME ... 2nd md5name case.
        sumstr = 'n,%s' % md5(bytes(os.path.basename(msg['relPath']),'utf-8')).hexdigest()

    if 'fileOp' in msg:
        if 'rename' in msg['fileOp']:
            msg['oldname'] = msg['fileOp']['rename']
        if 'link' in msg['fileOp']:
            hash = sha512()
            hash.update( bytes( msg['fileOp']['link'], encoding='utf-8' ) )
            sumstr = 'L,%s' % hash.hexdigest()
        elif 'remove' in msg['fileOp']:
            hash   = sha512()
            hash.update(bytes(os.path.basename(msg['relPath']), encoding='utf-8'))
            sumstr = 'R,%s' % hash.hexdigest()
        elif 'directory' in msg['fileOp']:
            hash   = sha512()
            hash.update(bytes(os.path.basename(msg['relPath']), encoding='utf-8'))

            if 'remove' in msg['fileOp']:
                sumstr = 'r,%s' % hash.hexdigest()
            else:
                sumstr = 'm,%s' % hash.hexdigest()

        else:
            logger.error('unknown fileOp: %s' % msg['fileOp'] )
    return sumstr

class Message:
    def __init__(self, h):
        """
         builds the in-memory representation of a message as expected by v2 plugins.
         In v3, a message is just a dictionary. in v2 it is an object.

         assign everything, except topic... because the topic is stored outside the body in v02.
        """

        self.pubtime = h['pubTime'].replace("T", "")
        self.baseurl = h['baseUrl']
        self.relpath = h['relPath']

        if 'new_dir' in h:
            self.new_dir = h['new_dir']
            self.new_file = h['new_file']

        if 'new_relPath' in h:
            self.new_relpath = h['new_relPath']

        self.urlstr = self.baseurl + self.relpath
        self.url = urllib.parse.urlparse(self.urlstr)

        self.notice = self.pubtime + ' ' + h["baseUrl"] + ' ' + h[
            "relPath"].replace(' ', '%20').replace('#', '%23')

        #FIXME: ensure headers are < 255 chars.
        for k in ['mtime', 'atime']:
            if k in h:
                h[k] = h[k].replace("T", "")

        #FIXME: sum header encoding.
        if 'size' in h:
            if type(h['size']) is str:
                h['size'] = int(h['size'])
            h['parts'] = '1,%d,1,0,0' % h['size']

        if 'blocks' in h:
            if h['blocks']['method'] == 'inplace':
                m = 'i'
            else:
                m = 'p'
            p = h['blocks']
            h['parts'] = '%s,%d,%d,%d,%d' % (m, p['size'], p['count'],
                                             p['remainder'], p['number'])

        h['topic'] = [ 'v02', 'post' ] + self.relpath.split('/')[0:-1]

        if 'parts' in h:
            self.partstr = h['parts']
        #else:
        #    self.partstr = None

        self.sumstr = sumstrFromMessage( h )
        self.sumflg = self.sumstr[0]
        h['sum'] = self.sumstr

        if 'fileOp' in h and  'rename' in h['fileOp'] :
            h['oldname'] = h['fileOp']['rename'] 
       
        self.headers = h
        self.hdrstr = str(h)
        self.isRetry = False

        # from sr_message/sr_new ...
        self.local_offset = 0
        self.in_partfile = False
        self.local_checksum = None

        self.target_file = None
        # does not cover partitioned files.

    def set_hdrstr(self):
        logger.info("set_hdrstr not implemented")
        pass

    def get_elapse(self):
        return nowflt() - timestr2flt(self.pubtime)

    def set_parts(self):
        logger.info("set_parts not implemented")
        pass


class V2Wrapper(FlowCB):
    def __init__(self, o):
        """
           A wrapper class to run v02 plugins.
           us run_entry(entry_point,module)

           entry_point is a string like 'on_message',  and module being the one to add.

           weird v2 stuff:   
                when calling init, self is a config/subscriber...
                when calling on_message, self is a message...
                that is kind of blown away for each message...
                parent is the config/subscriber in both cases.
                so v2 state variables are always stored in parent.

        """
        global logger

        logging.basicConfig(format=o.logFormat,
                            level=getattr(logging, o.logLevel.upper()))

        logger.setLevel(getattr(logging, o.logLevel.upper()))

        #logger.info('logging: fmt=%s, level=%s' % ( o.logFormat, o.logLevel ) )

        # FIXME, insert parent fields for v2 plugins to use here.
        self.logger = logger
        #logger.info('v2wrapper init start')

        self.state_vars = []

        if o.statehost:
            hostdir = o.hostdir
        else:
            hostdir = None

        self.user_cache_dir = sarracenia.config.get_user_cache_dir(hostdir)

        if hasattr(o, 'no'):
            self.instance = o.no
        else:
            self.instance = 0

        self.o = o

        self.v2plugins = {}
        self.consumer = types.SimpleNamespace()
        self.consumer.sleep_min = 0.01

        for ep in sarracenia.config.Config.v2entry_points:
            self.v2plugins[ep] = []

        unsupported_v2_events = ['do_download', 'do_get', 'do_put', 'do_send']
        for e in o.v2plugins:
            #logger.info('resolving: %s' % e)
            for v in o.v2plugins[e]:
                if e in unsupported_v2_events:
                    logger.error(
                        'v2 plugin conversion required, %s too different in v3'
                        % e)
                    continue
                self.add(e, v)

        #propagate options back to self.o for on_timing calls.
        #for v2o in self.o.v2plugin_options:
        #    setattr( self.o, v2o, getattr(self,v2o )  )

        # backward compat...
        self.o.user_cache_dir = self.o.cfg_run_dir
        self.o.instance = self.instance
        self.o.logger = self.logger
        if hasattr(self.o, 'post_baseDir'):
            self.o.post_base_dir = self.o.post_baseDir

        #logger.info('v2wrapper init done')

    def declare_option(self, option):
        logger.info('v2plugin option: %s declared' % option)

        self.state_vars.append(option)

        self.o.add_option(option)
        if not hasattr(self.o, option):
            logger.info('value of %s not set' % option)
            return

        if type(getattr(self.o, option)) is not list:
            setattr(self.o, option, [getattr(self.o, option)])

    def add(self, opname, path):

        setattr(self, opname, None)

        if path == 'None' or path == 'none' or path == 'off':
            logger.info("Reset plugin %s to None" % opname)
            exec('self.' + opname + '_list = [ ]')
            return True

        ok, script = sarracenia.config.config_path('plugins',
                                                   path,
                                                   mandatory=True,
                                                   ctype='py')
        if not ok:
            logger.error("installing %s %s failed: not found " %
                         (opname, path))
            return False

        #logger.info('installing: %s %s' % ( opname, path ) )

        c1 = set(vars(self))

        try:
            with open(script) as f:
                exec(
                    compile(f.read().replace('self.plugin', 'self.v2plugin'),
                            script, 'exec'))
        except:
            logger.error(
                "sr_config/execfile 2 failed for option '%s' and plugin '%s'" %
                (opname, path))
            logger.debug('Exception details: ', exc_info=True)
            return False

        if opname == 'plugin':
            if getattr(self, 'v2plugin') is None:
                logger.error("%s plugin %s incorrect: does not set self.%s" %
                             ('v2plugin', path, 'v2plugin'))
                return False

            # pci plugin-class-instance... parent is self (a v2wrapper)
            pci = self.v2plugin.lower()
            s = pci + ' = ' + self.v2plugin + '(self)'
            exec(pci + ' = ' + self.v2plugin + '(self)')
            s = 'vars(' + self.v2plugin + ')'
            pcv = eval('vars(' + self.v2plugin + ')')
            for when in sarracenia.config.Config.v2entry_points:
                if when in pcv:
                    #logger.info("v2 registering %s from %s" % ( when, path ) )

                    # 2020/05/22. I think the commented exec can be removed.
                    #FIXME: this breaks things horrible in v3. I do not see the usefulness even in v2.
                    #       everything is done with the lists, so value of setting individual value is nil.
                    #      self.on_start... vs.
                    #       self.v2plugins['on_start'].append( thing. )
                    #exec( 'self.' + when + '=' + pci + '.' + when )
                    eval('self.v2plugins["' + when + '"].append(' + pci + '.' +
                         when + ')')
        else:
            if getattr(self, opname) is None:
                logger.error("%s plugin %s incorrect: does not set self.%s" %
                             (opname, path, opname))
                return False

            #eval( 'self.' + opname + '_list.append(self.' + opname + ')' )
            eval('self.v2plugins["' + opname + '"].append( self.' + opname +
                 ')')

        c2 = set(vars(self))
        c12diff = list(c2 - c1)
        #logger.error('init added: +%s+ to %s' % (c12diff, self.state_vars) )
        if len(c12diff) > 0:
            self.state_vars.extend(c12diff)

        for opt in self.state_vars:
            if hasattr(self, opt):
                setattr(self.o, opt, getattr(self, opt))

        return True

    def after_work(self, worklist):
        ok_to_post = []
        for m in worklist.ok:
            if self.run_entry('on_file', m):
                ok_to_post.append(m)
            else:
                #worklist.failed.append(m)
                pass
                # FIXME: what should we do on failure of on_file plugin?
                #     download worked, but on_file failed... hmm...

        worklist.ok = ok_to_post

        outgoing = []
        for m in worklist.ok:
            if self.run_entry('on_post', m):
                outgoing.append(m)
            else:
                worklist.rejected.append(m)
        # set incoming for future steps.
        worklist.ok = outgoing

    def after_accept(self, worklist):

        outgoing = []
        for m in worklist.incoming:
            try:
                if self.run_entry('on_message', m):
                    outgoing.append(m)
                else:
                    worklist.rejected.append(m)
            except Exception as Ex:
               logger.error( f"plugin {m} died: {Ex}" );
               logger.debug( 'details: ', exc_info=True)
               worklist.rejected.append(m)
        # set incoming for future steps.
        worklist.incoming = outgoing

    def on_time(self, time):
        """
           run plugins for a given entry point.
        """
        logger.info('v2 run %s' % time)
        for plugin in self.v2plugins[time]:
            plugin(self.o)

    def on_housekeeping(self):
        self.on_time('on_heartbeat')

    def on_start(self):
        self.on_time('on_start')

    def on_stop(self):
        self.on_time('on_stop')

    def restoreMsg(self, m, v2msg):

        if 'topic' in m:
            if m['topic'][0:2] == ['v02', 'post' ]:
               m['topic'] = self.o.post_topicPrefix + m['topic'][2:]      

        if ('link' in v2msg.headers):
            if not 'fileOp' in m:
               m['fileOp'] = {}

            if m['fileOp']['link'] != v2msg.headers['link']:
                m['fileOp']['link'] = v2msg.headers['link']
            
        for h in ['oldname', 'newname' ]:
            if (h in v2msg.headers) and ((h not in m) or
                                         (v2msg.headers[h] != m[h])):
                m[h] = v2msg.headers[h]

        if v2msg.new_dir != m['new_dir']:
            m['new_dir'] = v2msg.new_dir
            relpath = m['new_dir'] + '/' + v2msg.new_file
            if self.o.post_baseDir:
                relpath = relpath.replace(self.o.post_baseDir, '', 1)
            m['new_relPath'] = relpath

        if v2msg.baseurl != m['baseUrl']:
            m['baseUrl'] = v2msg.baseurl

        if hasattr(v2msg, 'new_file'):
            if ('new_file' not in m) or (m['new_file'] != v2msg.new_file):
                m['new_file'] = v2msg.new_file

        if hasattr(
                v2msg,
                'post_base_dir') and (v2msg.post_base_dir != m['new_baseDir']):
            m['post_baseDir'] = v2msg.post_base_dir

    def run_entry(self, ep, m):
        """
           run plugins for a given entry point.
        """
        self.msg = Message(m)
        self.msg.topic = '.'.join(self.o.topicPrefix + m['subtopic'])
        self.o.msg = self.msg
        if hasattr(self.msg, 'partstr'):
            self.o.partstr = self.msg.partstr
        self.o.sumstr = self.msg.sumstr

        varsb4 = set(vars(self.msg))

        for opt in self.state_vars:
            if hasattr(self.o, opt):
                setattr(self.msg, opt, getattr(self.o, opt))

        ok = True
        for plugin in self.v2plugins[ep]:
            ok = plugin(self.o)
            if not ok: break

        vars_after = set(vars(self.msg))

        self.restoreMsg(m, self.msg)

        diff = list(vars_after - varsb4)
        if len(diff) > 0:
            self.state_vars.extend(diff)

        return ok



"""
    Adjust file paths, modifies the paths in file operation fileds, for file renaming,
    removal or links, or to give files different names on download.

    where files from a 'before' directory, are sent by an sr_sender to a 'after' directory.
    This also applies to the HPC mirroring case, where one would likely have a setting like

    pathReplace site5,site6
    pathReplaceFields dir

    sample usage:

    pathReplace before,after

    flowcb accept.pathreplace

"""

import logging

import os

from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class Pathreplace(FlowCB):
    def __init__(self, options):
        super().__init__(options, logger)

        self.o.add_option('pathReplace', 'list' )

        all_fields = set(['dir','file','rename','hlink','link'])
        self.o.add_option( 'pathReplaceFields', 'set', all_fields, all_fields )

    def on_start(self):
        if self.o.pathReplace == None:
            logger.error("pathReplace setting mandatory")
            return

        logger.debug("pathReplace is %s " % self.o.pathReplace )

    def after_accept(self, worklist):
        if self.o.pathReplace == None:
            return

        for msg in worklist.incoming:
            for p in self.o.pathReplace:
                (b, a) = p.split(",")

                # not sure if replacing in the main path is needed... maybe just in fileOp fields?
                logger.info("replace: %s by %s in: %s" % (b, a, self.o.pathReplaceFields) )

                if 'dir' in self.o.pathReplaceFields:
                    new_new_dir = msg['new_dir'].replace(b, a, 1)

                    if new_new_dir != msg['new_dir'] :
                        msg['new_dir'] = new_new_dir

                if 'file' in self.o.pathReplaceFields:
                    msg['new_file'] = msg['new_file'].replace(b, a, 1)

                msg['new_relPath'] = msg['new_dir'] + os.sep + msg['new_file']
    
                # adjust oldname/newname/link  if related to strings to replace
    
                if 'fileOp' in msg:
                    if 'rename' in msg['fileOp'] and 'rename' in self.o.pathReplaceFields:
                        msg['fileOp']['rename'] = msg['fileOp']['rename'].replace(b, a, 1)
                    if 'hlink' in msg['fileOp'] and 'hlink' in self.o.pathReplaceFields:
                        msg['fileOp']['hlink'] = msg['fileOp']['hlink'].replace(b, a, 1)
                    if 'link' in msg['fileOp'] and 'link' in self.o.pathReplaceFields:
                        msg['fileOp']['link'] = msg['fileOp']['link'].replace(b, a, 1)
    
                ## adjust new_relpath if posting    
                if self.o.post_baseDir:
                    msg['new_relPath'] = msg['new_relPath'].replace(self.o.post_baseDir, '',1)





"""

sarracenia.flowcb.accept modules are ones where the main focus is on the after_accept entry point.
This entry point is called after reject & accept rules have been called, and typically after
duplicate suppression has also been applied.

They can be used to further refine which files should be downloaded, by moving messages from the 
worklist.incoming to worklist.rejected.



"""

pass
